"""ASN.1 specification for X509 extensions."""

from smime.crypto.asn1 import named_value
from smime.crypto.asn1 import oid
from smime.crypto.asn1 import tag
from smime.crypto.asn1 import types
from smime.crypto.asn1 import x509_common
from smime.crypto.asn1 import x509_name


# Standard extensions from RFC 5280.
class BasicConstraints(types.Sequence):
    print_delimiter = ", "
    components = (
        (types.Component("cA", types.Boolean, default=False)),
        (types.Component("pathLenConstraint", types.Integer, optional=True))
        )


class SubjectAlternativeNames(types.SequenceOf):
    print_delimiter = ", "
    component = x509_name.GeneralName


class KeyUsage(types.NamedBitList):
    DIGITAL_SIGNATURE = named_value.NamedValue("digitalSignature", 0)
    NON_REPUDIATION = named_value.NamedValue("nonRepudiation", 1)
    KEY_ENCIPHERMENT = named_value.NamedValue("keyEncipherment", 2)
    DATA_ENCIPHERMENT = named_value.NamedValue("dataEncipherment", 3)
    KEY_AGREEMENT = named_value.NamedValue("keyAgreement", 4)
    KEY_CERT_SIGN = named_value.NamedValue("keyCertSign", 5)
    CRL_SIGN = named_value.NamedValue("cRLSign", 6)
    ENCIPHER_ONLY = named_value.NamedValue("encipherOnly", 7)
    DECIPHER_ONLY = named_value.NamedValue("decipherOnly", 8)
    named_bit_list = (DIGITAL_SIGNATURE, NON_REPUDIATION, KEY_ENCIPHERMENT,
                      DATA_ENCIPHERMENT, KEY_AGREEMENT, KEY_CERT_SIGN,
                      CRL_SIGN, ENCIPHER_ONLY, DECIPHER_ONLY)


class KeyPurposeID(oid.ObjectIdentifier):
    pass


class ExtendedKeyUsage(types.SequenceOf):
    print_delimiter = ", "
    print_labels = False
    component = KeyPurposeID


class KeyIdentifier(types.OctetString):
  pass


class SubjectKeyIdentifier(KeyIdentifier):
  pass


KEY_IDENTIFIER = "keyIdentifier"
AUTHORITY_CERT_ISSUER = "authorityCertIssuer"
AUTHORITY_CERT_SERIAL_NUMBER = "authorityCertSerialNumber"


class AuthorityKeyIdentifier(types.Sequence):
  components = (
      types.Component(KEY_IDENTIFIER, KeyIdentifier.implicit(0), optional=True),
      types.Component(AUTHORITY_CERT_ISSUER, x509_name.GeneralNames.implicit(1),
                      optional=True),
      types.Component(AUTHORITY_CERT_SERIAL_NUMBER,
                      x509_common.CertificateSerialNumber.implicit(2),
                      optional=True)
      )


class DisplayText(types.Choice):
    components = {
        "ia5String": types.IA5String,
        "visibleString": types.VisibleString,
        "bmpString": types.BMPString,
        "utf8String": types.UTF8String
        }


class NoticeNumbers(types.SequenceOf):
    component = types.Integer


class NoticeReference(types.Sequence):
    components = (
        types.Component("organization", DisplayText),
        types.Component("noticeNumbers", NoticeNumbers)
        )


NOTICE_REF = "noticeRef"
EXPLICIT_TEXT = "explicitText"


class UserNotice(types.Sequence):
    components = (
        types.Component(NOTICE_REF, NoticeReference, optional=True),
        types.Component(EXPLICIT_TEXT, DisplayText, optional=True)
)



class CPSuri(types.IA5String):
    pass


_POLICY_QUALIFIER_DICT = {
    oid.ID_QT_CPS: CPSuri,
    oid.ID_QT_UNOTICE: UserNotice
}


POLICY_QUALIFIER_ID = "policyQualifierId"
QUALIFIER = "qualifier"


class PolicyQualifierInfo(types.Sequence):
    print_labels = False
    print_delimiter = ": "
    components = (
        types.Component(POLICY_QUALIFIER_ID, oid.ObjectIdentifier),
        types.Component(QUALIFIER, types.Any, defined_by="policyQualifierId",
                        lookup=_POLICY_QUALIFIER_DICT)
        )


class PolicyQualifiers(types.SequenceOf):
    print_labels = False
    component = PolicyQualifierInfo


POLICY_IDENTIFIER = "policyIdentifier"
POLICY_QUALIFIERS = "policyQualifiers"


class PolicyInformation(types.Sequence):
    components = (
        types.Component(POLICY_IDENTIFIER, oid.ObjectIdentifier),
        types.Component(POLICY_QUALIFIERS, PolicyQualifiers, optional=True)
)

class CertificatePolicies(types.SequenceOf):
    component = PolicyInformation


FULL_NAME = "fullName"
RELATIVE_NAME = "nameRelativetoCRLIssuer"


class DistributionPointName(types.Choice):
    components = {
        FULL_NAME: x509_name.GeneralNames.implicit(0),
        RELATIVE_NAME: x509_name.RelativeDistinguishedName.implicit(1)
        }


class ReasonFlags(types.NamedBitList):
    UNUSED = named_value.NamedValue("unused", 0)
    KEY_COMPROMISE = named_value.NamedValue("keyCompromise", 1)
    CA_COMPROMISE = named_value.NamedValue("cACompromise", 2),
    AFFILIATION_CHANGED = named_value.NamedValue("affiliationChanged", 3)
    SUPERSEDED = named_value.NamedValue("superseded", 4)
    CESSATION_OF_OPERATION = named_value.NamedValue("cessationOfOperation", 5)
    CERTIFICATE_HOLD = named_value.NamedValue("certificateHold", 6)
    PRIVILEGE_WITHDRAWN = named_value.NamedValue("privilegeWithdrawn", 7)
    AA_COMPROMISE = named_value.NamedValue("aACompromise", 8)
    named_bit_list = (UNUSED, KEY_COMPROMISE, CA_COMPROMISE,
                      AFFILIATION_CHANGED, SUPERSEDED, CESSATION_OF_OPERATION,
                      CERTIFICATE_HOLD, PRIVILEGE_WITHDRAWN, AA_COMPROMISE)


DISTRIBUTION_POINT = "distributionPoint"
REASONS = "reasons"
CRL_ISSUER = "cRLIssuer"


class DistributionPoint(types.Sequence):
    components = (
        types.Component(DISTRIBUTION_POINT, DistributionPointName.explicit(0),
                        optional=True),
        types.Component(REASONS, ReasonFlags.implicit(1), optional=True),
        types.Component(CRL_ISSUER, x509_name.GeneralNames.implicit(2),
                        optional=True)
        )


class CRLDistributionPoints(types.SequenceOf):
    component = DistributionPoint


ACCESS_METHOD = "accessMethod"
ACCESS_LOCATION = "accessLocation"


class AccessDescription(types.Sequence):
    print_labels = False
    print_delimiter = ": "
    components = (
        types.Component(ACCESS_METHOD, oid.ObjectIdentifier),
        types.Component(ACCESS_LOCATION, x509_name.GeneralName)
)


# Called AuthorityInfoAccessSyntax in RFC 5280.
class AuthorityInfoAccess(types.SequenceOf):
    component = AccessDescription

class SignedCertificateTimestampList(types.OctetString):
    pass

# Hack! This is not a valid ASN.1 definition but it works: an extension value
# value is defined as a DER-encoded value wrapped in an OctetString.
# This is functionally equivalent to an Any type that is tagged with the
# OctetString tag.
@types.Universal(4, tag.PRIMITIVE)
class ExtensionValue(types.Any):
    pass


_EXTENSION_DICT = {
    oid.ID_CE_BASIC_CONSTRAINTS: BasicConstraints,
    oid.ID_CE_SUBJECT_ALT_NAME: SubjectAlternativeNames,
    oid.ID_CE_KEY_USAGE: KeyUsage,
    oid.ID_CE_EXT_KEY_USAGE: ExtendedKeyUsage,
    oid.ID_CE_SUBJECT_KEY_IDENTIFIER: SubjectKeyIdentifier,
    oid.ID_CE_AUTHORITY_KEY_IDENTIFIER: AuthorityKeyIdentifier,
    oid.ID_CE_CERTIFICATE_POLICIES: CertificatePolicies,
    oid.ID_CE_CRL_DISTRIBUTION_POINTS: CRLDistributionPoints,
    oid.ID_PE_AUTHORITY_INFO_ACCESS: AuthorityInfoAccess,
    oid.CT_POISON: types.Null,
    oid.CT_EMBEDDED_SCT_LIST: SignedCertificateTimestampList
    }


class Extension(types.Sequence):
    print_delimiter = ", "
    components = (
        types.Component("extnID", oid.ObjectIdentifier),
        types.Component("critical", types.Boolean, default=False),
        types.Component("extnValue", ExtensionValue, defined_by="extnID",
                         lookup=_EXTENSION_DICT)
        )


class Extensions(types.SequenceOf):
    component = Extension



#!/usr/bin/env python

import unittest

from smime.crypto import error
from smime.crypto.asn1 import tag
from smime.crypto.asn1 import types
from smime.crypto.asn1 import type_test_base


class TagDecoratorTest(unittest.TestCase):
    """Test the automatic creation of tags."""

    def test_universal_tag(self):
        class Test(object):
            tags = ()

        tagger = types.Universal(5, tag.PRIMITIVE)
        tagger(Test)

        self.assertEqual(1, len(Test.tags))
        expected_tag = tag.Tag(5, tag.UNIVERSAL, tag.PRIMITIVE)
        self.assertEqual(expected_tag, Test.tags[0])

    def test_explicit_tag(self):
        class Test(object):
            tags = ()

        tagger1 = types.Explicit(5, tag_class=tag.APPLICATION)
        tagger1(Test)

        self.assertEqual(1, len(Test.tags))
        expected_tag1 = tag.Tag(5, tag.APPLICATION, tag.CONSTRUCTED)
        self.assertEqual(expected_tag1, Test.tags[0])

        tagger2 = types.Explicit(3, tag_class=tag.CONTEXT_SPECIFIC)
        tagger2(Test)

        self.assertEqual(2, len(Test.tags))
        self.assertEqual(expected_tag1, Test.tags[0])

        expected_tag2 = tag.Tag(3, tag.CONTEXT_SPECIFIC, tag.CONSTRUCTED)
        self.assertEqual(expected_tag2, Test.tags[1])

    def test_implicit_tag(self):
        class Test(object):
            tags = ()

        tagger = types.Implicit(5, tag_class=tag.APPLICATION)
        # Cannot implicitly tag an untagged type.
        self.assertRaises(TypeError, tagger, Test)

        # Add a tag and try again.
        Test.tags = (tag.Tag(0, tag.UNIVERSAL, tag.PRIMITIVE),)
        tagger(Test)

        self.assertEqual(1, len(Test.tags))
        expected_tag = tag.Tag(5, tag.APPLICATION, tag.PRIMITIVE)
        self.assertEqual(expected_tag, Test.tags[0])

        # Repeat the test with a constructed encoding.
        Test.tags = (tag.Tag(0, tag.UNIVERSAL, tag.CONSTRUCTED),)
        tagger(Test)

        self.assertEqual(1, len(Test.tags))
        expected_tag = tag.Tag(5, tag.APPLICATION, tag.CONSTRUCTED)
        self.assertEqual(expected_tag, Test.tags[0])

# A dummy class we use to test that values are encoded as tag-length-value
# triplets.
class Dummy(types.Simple):
    # Fake.
    tags = (tag.Tag(1, tag.UNIVERSAL, tag.PRIMITIVE),)

    def _convert_value(cls, value):
        if isinstance(value, str):
            return value
        raise TypeError("Can't make a dummy from %s" % type(value))

    def _decode_value(self, buf, strict=True):
        return buf

    def _encode_value(self):
        return self._value

    def __str__(self):
        # Inject a marker to test human_readable().
        return "dummy!" + str(self._value)

# And a simple sequence to test some properties of constructe objects.
class DummySequence(types.Sequence):
    LOOK = {True: types.Integer}
    components = (
        types.Component("bool", types.Boolean),
        types.Component("int", types.Integer, optional=True),
        types.Component("oct", types.OctetString, default="hi"),
        types.Component("any", types.Any, defined_by="bool", lookup=LOOK)
      )


class TagLengthValueTest(unittest.TestCase):
    """Test Tag-Length-Value encoding."""
    def test_encode_decode_int(self):
        signed_integer_encodings = (
            (0, "00"),
            (127, "7f"),
            (128, "0080"),
            (256, "0100"),
            (-1, "ff"),
            (-128, "80"),
            (-129, "ff7f")
            )

        for value, enc in signed_integer_encodings:
            self.assertEqual(types.encode_int(value).encode("hex"), enc)
            self.assertEqual(types.decode_int(enc.decode("hex")), value)

        unsigned_integer_encodings = (
            (0, "00"),
            (127, "7f"),
            (128, "80"),
            (256, "0100")
            )

        for value, enc in unsigned_integer_encodings:
            self.assertEqual(
                types.encode_int(value, signed=False).encode("hex"), enc)
            self.assertEqual(
                types.decode_int(enc.decode("hex"), signed=False), value)

    def test_encode_read_length(self):
        length_encodings = (
            (0, "00"),
            (1, "01"),
            (38, "26"),
            (127, "7f"),
            (129, "8181"),
            (201, "81c9"),
            (65535, "82ffff"),
            (65536, "83010000")
            )

        for value, enc in length_encodings:
            self.assertEqual(types.encode_length(value).encode("hex"), enc)
            self.assertEqual(types.read_length(enc.decode("hex")), (value, ""))
            # Test that the reader stops after the specified number of bytes.
            longer = enc + "00"
            self.assertEqual(types.read_length(longer.decode("hex")),
                             (value, "\x00"))
            longer = enc + "ff"
            self.assertEqual(types.read_length(longer.decode("hex")),
                             (value, "\xff"))
            # And test that it complains when there are not enough bytes.
            shorter = enc[:-2]
            self.assertRaises(error.ASN1Error,
                              types.read_length, shorter.decode("hex"))

    def test_read_indefinite_length(self):
        indef_length = "80".decode("hex")
        self.assertRaises(error.ASN1Error, types.read_length, indef_length)
        self.assertEqual(types.read_length(indef_length, strict=False),
                         (-1, ""))

        self.assertEqual(types.read_length(indef_length + "hello", strict=False),
                         (-1, "hello"))

    def test_encode_decode_read(self):
        value = "hello"
        d = Dummy(value=value)
        enc = d.encode()
        encoded_length = types.encode_length(len(value))

        expected = Dummy.tags[0].value + encoded_length + value
        self.assertEqual(expected.encode("hex"), enc.encode("hex"))

        decoded_dummy = Dummy.decode(enc)
        self.assertTrue(isinstance(decoded_dummy, Dummy))
        self.assertEqual(decoded_dummy.value, value)

        read_dummy, rest = Dummy.read(enc)
        self.assertTrue(isinstance(read_dummy, Dummy))
        self.assertEqual(read_dummy.value, value)
        self.assertEqual("", rest)

    def test_read_from_beginning(self):
        value = "hello"
        d = Dummy(value=value)
        self.assertEqual("hello", d.value)
        enc = d.encode()

        encoded_length = types.encode_length(len(d.value))
        expected = Dummy.tags[0].value + encoded_length + d.value
        self.assertEqual(expected.encode("hex"), enc.encode("hex"))

        longer_buffer = enc + "ello"
        # We can't decode because there are leftover bytes...
        self.assertRaises(error.ASN1Error, Dummy.decode, longer_buffer)

        # ... but we can read from the beginning of the buffer.
        read_dummy, rest = Dummy.read(longer_buffer)
        self.assertTrue(isinstance(read_dummy, Dummy))
        self.assertEqual("hello", read_dummy.value)
        self.assertEqual("ello", rest)

    def test_encode_decode_read_multiple_tags(self):
        @types.Explicit(8)
        class NewDummy(Dummy):
            pass

        value = "hello"
        d = NewDummy(value=value)
        enc = d.encode()
        encoded_inner_length = types.encode_length(len(value))

        inner = Dummy.tags[0].value + encoded_inner_length + value
        encoded_length = types.encode_length(len(inner))

        expected = NewDummy.tags[1].value + encoded_length + inner
        self.assertEqual(expected.encode("hex"), enc.encode("hex"))

        decoded_dummy = NewDummy.decode(enc)
        self.assertTrue(isinstance(decoded_dummy, NewDummy))
        self.assertEqual(decoded_dummy.value, value)

        read_dummy, rest = NewDummy.read(enc)
        self.assertTrue(isinstance(read_dummy, NewDummy))
        self.assertEqual(read_dummy.value, value)
        self.assertEqual("", rest)

        indef_encoding = "a880010568656c6c6f0000".decode("hex")
        self.assertRaises(error.ASN1Error, NewDummy.decode, indef_encoding)
        self.assertEqual(NewDummy.decode(indef_encoding, strict=False),
                         NewDummy(value="hello"))


class BooleanTest(type_test_base.TypeTestBase):
    asn1_type = types.Boolean
    repeated = False
    keyed = False
    initializers = (
        (False, 0),
        (True, 1),
        )
    bad_initializers = (
      # Everything is converted to a bool and accepted.
      )
    encode_test_vectors = (
        (True, "0101ff"),
        (False, "010100")
        )
    bad_encodings = (
        # Empty value.
        ("0100"),
        # Longer than 1 byte.
        ("01020000"),
        ("0102ffff"),
        # Indefinite length
        ("0180ff0000")
        )
    bad_strict_encodings = (
        # Nonzero byte for True.
        (True, "010101"),
        (True, "0101ab")
        )


class IntegerTest(type_test_base.TypeTestBase):
    asn1_type = types.Integer
    repeated = False
    keyed = False
    initializers = (
        (0,),
        (1,),
        (-1,),
        (1000000,),
        )
    bad_initializers = (
        # Everything that can be converted to an int is accepted.
        )
    encode_test_vectors = (
        (0, "020100"),
        (127, "02017f"),
        (128, "02020080"),
        (256, "02020100"),
        (-1, "0201ff"),
        (-128, "020180"),
        (-129, "0202ff7f")
        )
    bad_encodings = (
        # Empty value.
        ("0200"),
        # Indefinite length.
        ("0280ff0000")
        )
    bad_strict_encodings = (
        # Leading 0-octets.
        (0, "02020000"),
        (127, "0202007f"),
        # Leading ff-octets.
        (-1, "0202ffff"),
        (-128, "0202ff80")
      )


class OctetStringTest(type_test_base.TypeTestBase):
    asn1_type = types.OctetString
    repeated = False
    keyed = False
    initializers = (
        ("hello",),
        ("\xff\x00",),
        )
    bad_initializers = (
        # Nothing exciting.
        )
    encode_test_vectors = (
        # Empty strings are allowed.
        ("", "0400"),
        ("hello", "040568656c6c6f"),
        ("\xff\x00", "0402ff00")
        )
    bad_encodings = (
      # Indefinite length.
      ("0480abcdef0000"),
      )
    bad_strict_encodings = ()

# Skip other string type tests as there's currently no exciting specialization
# for those.


class BitStringTest(type_test_base.TypeTestBase):
    asn1_type = types.BitString
    repeated = False
    keyed = False
    initializers = (
        ("",),
        ("0",),
        ("1",),
        ("010100010110",),
        )
    bad_initializers = (
        ("hello", ValueError),
        ("0123cdef", ValueError),
        ("\xff\x00", ValueError)
        )
    encode_test_vectors = (
        # From the ASN.1 spec.
        # 0a3b5f291cd
        ("00001010001110110101111100101001000111001101", "0307040a3b5f291cd0"),
        # More test vectors with different amounts of padding
        ("", "030100"),
        ("0", "03020700"),
        ("1", "03020780"),
        ("0000000", "03020100"),
        ("0000001", "03020102"),
        ("1000000", "03020180"),
        ("00000000", "03020000"),
        ("11111111", "030200ff"),
        ("0000000001", "0303060040"),
        )
    bad_encodings = (
        # Empty value - padding byte must always be present.
        ("0300"),
        # Padding but no other bytes.
        ("030101"),
        ("030107"),
        # Invalid padding value.
        ("030108"),
        ("030180"),
        ("03020800"),
        ("03028000"),
        # Invalid padding bits.
        ("030201ff"),
        ("030205f0"),
        ("030207f0"),
        # Indefinite length.
        ("038007800000")
        )
    bad_strict_encodings = ()


# Mix-in from object so the tests are not run for the base class itself.
class RepeatedTest(object):
    def test_modify_repeated(self):
        d = Dummy(value="world")
        d2 = Dummy(value="hello")
        s = self.asn1_type(value=[d])
        self.assertFalse(s.modified())
        original_enc = s.encode()

        s[0] = d2
        self.assertTrue(s.modified())
        self.assertEqual(s, [d2])
        self.assertNotEqual(s.encode(), original_enc)

        del s[0]
        self.assertTrue(s.modified())
        self.assertFalse(list(s))
        self.assertNotEqual(s.encode(), original_enc)

        # Back to original; but the modified bit is never cleared.
        s.append(d)
        self.assertTrue(s.modified())
        self.assertEqual(s, [d])
        self.assertEqual(s.encode(), original_enc)


class SequenceOfTest(type_test_base.TypeTestBase, RepeatedTest):
    # Test with a dummy class.
    class SequenceOfDummies(types.SequenceOf):
        component = Dummy
    asn1_type = SequenceOfDummies
    immutable = False
    keyed = False
    initializers = (
        ([Dummy(value="world"), Dummy(value="hello"), Dummy(value="\x00")],
         ["world", "hello", "\x00"],
         [Dummy(value="world"), "hello", "\x00"]),
        ([], ()),
      )
    bad_initializers = (
        # Can't coerce to Dummy.
        ([3], TypeError),
        ([True], TypeError),
        # Can't iterate.
        (True, TypeError)
        )
    encode_test_vectors = (
        ([], "3000"),
        ([Dummy(value="hello"), Dummy(value="\x00\xff")],
         "300b010568656c6c6f010200ff"),
        # Different order produces a different encoding.
        ([Dummy(value="\x00\xff"), Dummy(value="hello")],
         "300b010200ff010568656c6c6f")
        )
    bad_encodings = (
        # Bad element length.
        "3003010200",
        # Bad component tag.
        "30020200",
        # Indef length with no EOC.
        "3080010568656c0000010200ff",
        )

    bad_strict_encodings = ()

    def test_indefinite_length_encoding(self):
        # We cannot use bad_strict_encodings because of the re-encoding bug:
        # indefinite length is not preserved.
        # For good measure, we add an EOC in the contents.
        value = self.asn1_type([Dummy(value="hel\x00\x00"),
                                Dummy(value="\x00\xff")])
        indef_length_encoding = "3080010568656c0000010200ff0000".decode("hex")
        self.assertRaises(error.ASN1Error,
                          self.asn1_type.decode, indef_length_encoding)
        o = self.asn1_type.decode(indef_length_encoding, strict=False)
        self.assertEqual(o, value)


class SetOfTest(type_test_base.TypeTestBase, RepeatedTest):
    class SetOfDummies(types.SetOf):
        component = Dummy
    asn1_type = SetOfDummies
    immutable = False
    keyed = False
    initializers = (
        ([Dummy(value="world"), Dummy(value="\x00"), Dummy(value="world")],
         ["world", "\x00", "world"],
         [Dummy(value="world"), "\x00", "world"]),
        ([], ()),
      )
    bad_initializers = (
        # Can't coerce to Dummy.
        ([3], TypeError),
        ([True], TypeError),
        # Can't iterate.
        (True, TypeError)
        )
    encode_test_vectors = (
        ([], "3100"),
        # Elements are sorted according to their encoding.
        ([Dummy(value="\x00\xff"), Dummy(value="hello")],
         "310b010200ff010568656c6c6f"),
        )
    bad_encodings = (
        # Bad element length.
        "31010200",
        # Bad component tag.
        "31020200",
        # Indef length with no EOC.
        "3180010568656c0000010200ff",
        )
    bad_strict_encodings = (
        )

    def test_encoding_is_order_independent(self):
        elems = [Dummy(value="world"), Dummy(value="hello")]
        dummies = self.asn1_type(elems)
        elems2 = [Dummy(value="hello"), Dummy(value="world")]
        dummies2 = self.asn1_type(elems2)
        # Encodings compare equal even though the sets don't.
        self.assertEqual(dummies.encode(), dummies2.encode())

    def test_indefinite_length_encoding(self):
        # We cannot use bad_strict_encodings because of the re-encoding bug:
        # indefinite length is not preserved.
        # For good measure, we add an EOC in the contents.
        value = self.asn1_type([Dummy(value="hel\x00\x00"),
                                Dummy(value="\x00\xff")])
        indef_length_encoding = "3180010568656c0000010200ff0000".decode("hex")
        self.assertRaises(error.ASN1Error,
                          self.asn1_type.decode, indef_length_encoding)
        o = self.asn1_type.decode(indef_length_encoding, strict=False)
        self.assertEqual(o, value)


class AnyTest(type_test_base.TypeTestBase):
    asn1_type = types.Any
    repeated = False
    keyed = False
    initializers = (
        # Decoded and undecoded initializers.
        # Test with a few simple types.
        (types.Boolean(value=True).encode(), types.Boolean(value=True)),
        (types.Integer(value=3).encode(), types.Integer(value=3)),
        (types.OctetString("hello").encode(), types.OctetString("hello")),
        # We don't currently check that the encoded value encodes a valid
        # tag-length-value triplet, so this will also succeed,
        ("0000ff",),
        ("",)
        )
    bad_initializers = (
        (types.Any("hello"), TypeError),
        )
    encode_test_vectors = (
        # A Boolean True.
        ("\x01\x01\xff", "0101ff"),
        # An Integer 3.
        ("\x02\x01\x03", "020103"),
        # An octet string "hello".
        ("\x04\x05\x68\x65\x6c\x6c\x6f", "040568656c6c6f"),
        )
    bad_encodings = ()
    bad_strict_encodings = ()

    def test_decode_inner(self):
        dummy = Dummy(value="hello")
        a = types.Any(dummy)
        self.assertTrue(a.decoded)
        self.assertEqual(a.decoded_value, dummy)
        enc = dummy.encode()
        a2 = types.Any(enc)
        self.assertFalse(a2.decoded)
        self.assertEqual(a, a2)
        self.assertEqual(a.value, a2.value)
        a2.decode_inner(value_type=Dummy)
        self.assertTrue(a2.decoded)
        self.assertEqual(a2.decoded_value, dummy)

    def test_modify(self):
        dummy = DummySequence(value={"bool": True, "any": "\x01\x01\xff"})
        a = types.Any(dummy)
        self.assertFalse(a.modified())
        a.decoded_value["bool"] = False
        self.assertTrue(a.modified())

        enc = dummy.encode()
        a2 = types.Any(enc)
        self.assertFalse(a2.modified())
        a2.decode_inner(value_type=DummySequence)
        self.assertFalse(a2.modified())
        a2.decoded_value["bool"] = True
        self.assertTrue(a2.modified())


class ChoiceTest(type_test_base.TypeTestBase):
    class MyChoice(types.Choice):
        components = {
            "bool": types.Boolean,
            "int": types.Integer,
            "oct": types.OctetString,
          }
    asn1_type = MyChoice
    immutable = False
    repeated = False
    keyed = True
    initializers = (
        ({"bool": types.Boolean(value=False)}, {"bool": False}),
        ({"int": types.Integer(value=3)}, {"int": 3}),
        ({"oct": types.OctetString(value="hello")}, {"oct": "hello"}),
        ({}, {"bool": None}, {"int": None}, {"oct": None})
        )
    bad_initializers = (
        # Multiple values set at once.
        ({"bool": False, "int": 3}, ValueError),
        # Invalid key.
        ({"boo": False}, ValueError),
        )
    encode_test_vectors = (
        ({"bool": True}, "0101ff"),
        ({"int": 3}, "020103"),
        ({"oct": "hello"}, "040568656c6c6f"),
        )
    bad_encodings = ()
    bad_strict_encodings = ()

    def test_modify(self):
        m = self.MyChoice(value={"bool": True})
        self.assertFalse(m.modified())

        m["bool"] = False
        self.assertTrue(m.modified())
        self.assertFalse(m["bool"])

        # Back to original; but the modified bit is never cleared.
        m["bool"] = True
        self.assertTrue(m.modified())
        self.assertTrue(m["bool"])


class SequenceTest(type_test_base.TypeTestBase):
    asn1_type = DummySequence
    immutable = False
    repeated = False
    keyed = True
    initializers = (
        # Fully specified, Any can be decoded.
        ({"bool": True, "int": 3, "oct": "hello", "any": "\x02\x01\x05"},),
        # Fully specified, Any cannot be decoded.
        ({"bool": False, "int": 3, "oct": "hello", "any": "\x02\x01\x05"},),
        # Partially specified.
        ({"bool": True, "int": None, "oct": "hi", "any": None},
         {"bool": True},),
        ({"bool": None, "int": 3, "oct": "hi", "any": None}, {"int": 3},),
        # Setting the defaults is the same as setting nothing.
        ({"bool": None, "int": None, "oct": "hi", "any": None},
         {"bool": None, "int": None, "oct": None, "any": None},
         {},
         {"oct": "hi"}),
        )
    bad_initializers = (
        # Invalid key.
        ({"boo": False}, ValueError),
        # Invalid component.
        ({"int": "hello"}, ValueError)
        )
    encode_test_vectors = (
        ({"bool": True, "int": 3, "oct": "hello", "any": "\x02\x01\x05"},
         "30100101ff020103040568656c6c6f020105"),
        # Missing optional.
        ({"bool": True, "oct": "hello", "any": "\x02\x01\x05"},
         "300d0101ff040568656c6c6f020105"),
        # Missing default.
        ({"bool": True, "int": 3, "any": "\x02\x01\x05"},
         "30090101ff020103020105"),
        # Default value set.
        ({"bool": True, "int": 3, "oct": "hi", "any": "\x02\x01\x05"},
         "30090101ff020103020105"),
        )
    bad_encodings = (
        # Indef length with no EOC.
        "30800101ff020103040568656c0000020105",
        )
    bad_strict_encodings = ()

    def test_modify(self):
        s = DummySequence(value={"bool": True, "int": 2})
        self.assertFalse(s.modified())

        s["bool"] = False
        self.assertTrue(s.modified())
        self.assertFalse(s["bool"])
        self.assertEqual(s["int"], 2)

        # Back to original; but the modified bit is never cleared.
        s["bool"] = True
        self.assertTrue(s.modified())
        self.assertTrue(s["bool"])
        self.assertEqual(s["int"], 2)

    def test_decode_any(self):
        seq = self.asn1_type({"bool": True, "int": 3, "oct": "hello",
                              "any": "\x02\x01\x05"})
        enc = seq.encode()
        dec = self.asn1_type.decode(enc)
        self.assertTrue(dec["any"].decoded)
        self.assertEqual(dec["any"].decoded_value, 5)

        # Lookup key not in dictionary.
        seq = self.asn1_type({"bool": False, "int": 3, "oct": "hello",
                              "any": "\x02\x01\x05"})
        enc = seq.encode()
        seq = self.asn1_type.decode(enc)
        self.assertFalse(seq["any"].decoded)

        # Corrupt any.
        # We don't currently verify the Any spec when creating an element.
        seq = self.asn1_type({"bool": True, "int": 3, "oct": "hello",
                              "any": "\x01\x01\x05"})
        enc = seq.encode()
        # Can't decode in strict mode.
        self.assertRaises(error.ASN1Error, self.asn1_type.decode, enc)
        dec = self.asn1_type.decode(enc, strict=False)
        self.assertFalse(dec["any"].decoded)

    def test_indefinite_length_encoding(self):
        # We cannot use bad_strict_encodings because of the re-encoding bug:
        # indefinite length is not preserved.
        # For good measure, we add an EOC in the contents.
        value = self.asn1_type(value={"bool": True, "int": 3,
                                      "oct": "hel\x00\x00",
                                      "any": "\x02\x01\x05"})
        indef_length_encoding = (
            "30800101ff020103040568656c00000201050000".decode("hex"))
        self.assertRaises(error.ASN1Error,
                          self.asn1_type.decode, indef_length_encoding)
        o = self.asn1_type.decode(indef_length_encoding, strict=False)
        self.assertEqual(o, value)


# Some attempted test coverage for recursive mutable types.
class RecursiveTest(type_test_base.TypeTestBase):
    class SequenceOfSequence(types.SequenceOf):
        component = DummySequence
    asn1_type = SequenceOfSequence
    immutable = False
    repeated = True
    keyed = False
    initializers = (
        # Fully specified sequence.
        ([{"bool": True, "int": 3, "oct": "hello", "any": "\x02\x01\x05"}],),
        # Partially specified sequence.
        ([{"bool": True, "int": None, "oct": "hi", "any": None}],
         [{"bool": True}],),
        # Empty sequence.
        ([],)
        )
    bad_initializers = (
        # Invalid key in component.
        ([{"boo": False}], ValueError),
        # Invalid value in component.
        ([{"int": "hello"}], ValueError),
        # Invalid component: not iterable.
        (types.Boolean(True), TypeError),
        # Invalid component: iterable but wrong components.
        ([types.Boolean(True)], TypeError)
        )
    encode_test_vectors = (
        ([{"bool": True, "int": 3, "oct": "hello", "any": "\x02\x01\x05"}],
         "301230100101ff020103040568656c6c6f020105"),
        )
    bad_encodings = ()
    bad_strict_encodings = ()

    def test_modify_recursively(self):
        d = DummySequence(value={"bool": True, "int":3, "any": "\x02\x01\x05"})
        s = self.SequenceOfSequence(value=[d])
        self.assertFalse(s.modified())
        original_enc = s.encode()

        # Modify subcomponent.
        s[0]["bool"] = False
        self.assertTrue(s.modified())
        self.assertNotEqual(s.encode(), original_enc)

        # Reset.
        s[0]["bool"] = True
        self.assertTrue(s.modified())
        self.assertEqual(s.encode(), original_enc)


class PrintTest(unittest.TestCase):
    def test_simple_human_readable(self):
        dummy = Dummy("hello")
        # Ensure there's some content.
        self.assertTrue(str(dummy))
        self.assertTrue(str(dummy) in dummy.human_readable(wrap=0))

    def test_simple_human_readable_prints_label(self):
        s = Dummy("hello").human_readable(label="world")
        self.assertTrue("world" in s)

    def test_simple_human_readable_lines_wrap(self):
        dummy = Dummy(value="hello")
        wrap = 3
        for line in dummy.human_readable_lines(wrap=wrap):
            self.assertTrue(len(line) <= wrap)

    def test_string_value_int(self):
        i = types.Integer(value=123456789)
        self.assertTrue("123456789" in str(i))

    def test_string_value_bool(self):
        b = types.Boolean(value=True)
        self.assertTrue("true" in str(b).lower())
        b = types.Boolean(value=False)
        self.assertTrue("false" in str(b).lower())

    def test_string_value_string(self):
        # Currently all string types are just str, with no encoding.
        hello = "\x68\x65\x6c\x6c\x6f"
        invalid_printable_char = "*"
        opaque = "\xd7\xa9\xd7\x9c\xd7\x95\xd7\x9d"
        string_types = [types.TeletexString, types.PrintableString,
                        types.UniversalString, types.UTF8String,
                        types.BMPString, types.IA5String,
                        types.VisibleString]
        
        should_fail = {
            hello: [],
            invalid_printable_char: [types.PrintableString],
            opaque: [types.PrintableString,
                     types.IA5String,
                     types.VisibleString],
        }
        strings = [hello, invalid_printable_char, opaque]

        for t in string_types:
            # TODO(laiqu) make this fail for strings other than printable, ia5
            # and visible (and possibly make more specific character sets for
            # ia5/visible).
            for str_ in strings:
                if t not in should_fail[str_]:
                    s = t(serialized_value=str_, strict=True)
                    self.assertTrue(str_ in str(s))
                else:
                    self.assertRaises(error.ASN1Error, t,
                                      serialized_value=str_, strict=True)

    def test_string_value_bitstring(self):
        # 0x1ae
        b = str(types.BitString(value="0110101110"))
        self.assertTrue("1" in b)
        self.assertTrue("ae" in b.lower())

    def test_string_value_octetstring(self):
        b = str(types.OctetString(value="\x42\xac"))
        self.assertTrue("42" in b)
        self.assertTrue("ac" in b.lower())

    def test_constructed_human_readable(self):
        dummy = DummySequence({"bool": True, "int": 3})
        s = dummy.human_readable(wrap=0)
        self.assertTrue("bool" in s)
        self.assertTrue("true" in s.lower())
        self.assertTrue("int" in s)
        self.assertTrue("3" in s)
        # Present since a default is set.
        self.assertTrue("oct" in s)
        # Not present and no default.
        self.assertFalse("any" in s)


if __name__ == '__main__':
    unittest.main()



import unittest


from smime.crypto import error


class TypeTestBase(unittest.TestCase):
    # Test class for each concrete type should fill this in.
    asn1_type = None
    # Immutable types support hashing.
    immutable = True
    # Repeated types support lookup and assignment by index.
    repeated = True
    # Keyed types support lookup and assignment by key.
    keyed = True
    # A tuple of initializer tuples; components in each tuple should yield
    # equal objects. The first component in each tuple should be the canonical
    # value (returned by .value).
    initializers = None
    # A tuple of (bad_initializer, exception_raised) pairs.
    bad_initializers = None
    # A tuple of (value, hex_der_encoding) pairs.
    # Note: test vectors should include the complete encoding (including tag
    # and length). This is so we can lift test vectors directly from the ASN.1
    # spec and test that we recognize the correct tag for each type.
    # However test vectors for invalid encodings should focus on type-specific
    # corner cases. It's not necessary for each type to verify that invalid
    # tags and lengths are rejected: this is covered in separate tests.
    encode_test_vectors = None
    # A tuple of of serialized, hex-encoded values.
    bad_encodings = None
    # A tuple of (value, hex_encoding) pairs that can only be decoded
    # in non-strict mode.
    bad_strict_encodings = None

    def test_create(self):
        for initializer_set in self.initializers:
            value = initializer_set[0]
            # The canonical initializer.
            for i in initializer_set:
                o1 = self.asn1_type(value=i)
                self.assertEqual(o1.value, value)
                # And other initializers that yield the same value.
                for j in initializer_set:
                    o2 = self.asn1_type(value=j)
                    self.assertEqual(o2.value, value)
                    self.assertEqual(o1, o2)
                    if self.immutable:
                        self.assertEqual(hash(o1), hash(o2))
                    elif self.repeated:
                        self.assertEqual(len(o1), len(o2))
                        for i in range(len(o1)):
                            self.assertEqual(o1[i], o2[i])
                    elif self.keyed:
                        self.assertEqual(len(o1), len(o2))
                        self.assertEqual(o1.keys(), o2.keys())
                        for key in o1:
                            self.assertEqual(o1[key], o2[key])

        # Sanity-check: different initializers yield different values.
        for i in range(len(self.initializers)):
            for j in range(i+1, len(self.initializers)):
                o1 = self.asn1_type(value=self.initializers[i][0])
                o2 = self.asn1_type(value=self.initializers[j][0])
                self.assertNotEqual(o1, o2)
                if self.immutable:
                    self.assertNotEqual(hash(o1), hash(o2))
                self.assertNotEqual(o1.value, o2.value)

    def test_create_fails(self):
        for init, err in self.bad_initializers:
            self.assertRaises(err, self.asn1_type, init)

    def test_encode_decode(self):
        for value, enc in self.encode_test_vectors:
            o1 = self.asn1_type(value=value)
            o2 = self.asn1_type.decode(enc.decode("hex"))
            self.assertEqual(o1, o2)
            self.assertEqual(o1.value, o2.value)
            self.assertEqual(enc, o1.encode().encode("hex"))
            self.assertEqual(enc, o2.encode().encode("hex"))

    def test_decode_fails(self):
        for bad_enc in self.bad_encodings:
            self.assertRaises(error.ASN1Error, self.asn1_type.decode,
                bad_enc.decode("hex"))
            self.assertRaises(error.ASN1Error, self.asn1_type.decode,
                bad_enc.decode("hex"), strict=False)

    def test_strict_decode_fails(self):
        for value, bad_enc in self.bad_strict_encodings:
            o = self.asn1_type(value=value)
            self.assertRaises(error.ASN1Error,
                              self.asn1_type.decode, bad_enc.decode("hex"))
            o2 = self.asn1_type.decode(bad_enc.decode("hex"), strict=False)
            self.assertEqual(o, o2)
            # The object should keep its original encoding...
            self.assertEqual(bad_enc, o2.encode().encode("hex"))
            # ... which is not the canonical encoding.
            self.assertNotEqual(bad_enc, o.encode().encode("hex"))



"""Read and write PEM files and strings."""

import base64
import StringIO

from smime.crypto import error
import types


class PemError(error.EncodingError):
    pass

_START_TEMPLATE = "-----BEGIN %s-----"
_END_TEMPLATE = "-----END %s-----"


class PemReader(object):
    """A reader class for iteratively reading PEM files."""

    def __init__(self, fileobj, markers, skip_invalid_blobs=True):
        """Create a PemReader from a file object.

        When used as a context manager, the file object is closed
        upon exit.

        Args:
            fileobj: the file object to read from.
            markers: a single string or an iterable of markers accepted by the
                reader, e.g., CERTIFICATE, RSA PUBLIC KEY, etc.
            skip_invalid_blobs: if False, invalid PEM blobs cause a PemError.
                If True, invalid blobs are skipped. In non-skip mode, an
                immediate StopIteration before any valid blocks are found, also
                causes a PemError exception.

        Raises:
            PemError: invalid PEM contents.
        """
        if type(markers) in types.StringTypes:
            markers = (markers,)
        self.__f = fileobj
        self.__marker_dict = ({(_START_TEMPLATE % m): m for m in markers})
        self.__valid_blobs_read = 0
        self.__eof = False
        self.__skip_invalid_blobs = skip_invalid_blobs

    def __iter__(self):
        """Iterate over file contents.

        Returns:
            a generator function that yields decoded (blob, marker)
                tuples.
        """
        return self.read_blocks()

    def close(self):
        """Close the underlying file object."""
        self.__f.close()

    def __enter__(self):
        return self

    def __exit__(self, unused_type, unused_value, traceback):
        self.close()

    @classmethod
    def from_file(cls, pem_file, markers, skip_invalid_blobs=True):
        """Create a PemReader for reading a file.

        Caller is responsible for closing the reader afterwards.

        Args:
            pem_file: the file to read from.
            markers: an iterable of markers accepted by the reader,e.g.,
                CERTIFICATE, RSA PUBLIC KEY, etc.
            skip_invalid_blobs: if False, invalid PEM blobs cause a PemError.
                If True, invalid blobs are skipped. In non-skip mode, an
                immediate StopIteration before any valid blocks are found, also
                causes a PemError exception.

        Returns:
            a PemReader object.

        Raises:
            IOError, ValueError: the fileobject could not be operated on.
        """
        return cls(open(pem_file, "r"), markers, skip_invalid_blobs)

    @classmethod
    def from_string(cls, pem_string, markers, skip_invalid_blobs=True):
        """Create a PemReader for reading a string.

        Args:
            pem_string: the string to read from.
            markers: an iterable of markers accepted by the reader, e.g.,
                CERTIFICATE, RSA PUBLIC KEY, etc.
            skip_invalid_blobs: if False, invalid PEM blobs cause a PemError.
                If True, invalid blobs are skipped. In non-skip mode, an
                immediate StopIteration before any valid blocks are found, also
                causes a PemError exception.

        Returns:
            a PemReader object.
        """
        f = StringIO.StringIO(pem_string)
        return cls(f, markers, skip_invalid_blobs)

    def read_blocks(self):
        """Read the next PEM blob.

        Yields:
            (raw_string, marker) tuples containing the decoded blob and the
            marker used to detect the blob.

        Raises:
            PemError: a PEM block was invalid (in skip_invalid_blobs mode).
            IOError, ValueError: the file object could not be operated on.
            StopIteration: EOF was reached.
        """
        while not self.__eof:
            marker = None
            for line in self.__f:
                line = line.rstrip("\r\n")
                # PEM (RFC 1421) allows arbitrary comments between PEM blocks
                # so we skip over those
                if line in self.__marker_dict:
                    marker = self.__marker_dict[line]
                    break

            if not marker:
                self.__eof = True
                if (not self.__skip_invalid_blobs and
                    not self.__valid_blobs_read):
                    raise PemError("No PEM header")
                raise StopIteration

            ret = ""
            footer = _END_TEMPLATE % marker
            footer_found = False

            for line in self.__f:
                line = line.rstrip("\r\n")
                if line == footer:
                    footer_found = True
                    break
                ret += line

            # Here, we assume that each header is exactly matched by a footer.
            # TODO(ekasper): determine if this assumption is overly strict,
            # i.e., whether blocks such as BEGIN RSA PUBLIC KEY...END PUBLIC KEY
            # are commonly used in applications.
            if not footer_found:
                self.__eof = True
                if not self.__skip_invalid_blobs:
                    raise PemError("No PEM footer line to match the header")
                raise StopIteration

            try:
                # We don't use ret.decode('base64') here as the exceptions from
                # this method are not properly documented.
                yield base64.b64decode(ret), marker
                self.__valid_blobs_read += 1
            except TypeError:
                if not self.__skip_invalid_blobs:
                    # We do not set EOF here so caller can resume - even though
                    # this can normally be transparently handled by setting
                    # skip_invalid_blobs to True upon init.
                    raise PemError("Invalid base64 encoding")
                # Else just continue the loop
        raise StopIteration


class PemWriter(object):
    """A class for writing PEM blobs."""

    def __init__(self, fileobj, marker):
        """Create a writer.

        When used as a context manager, the underlying file object is closed
        upon exit.

        Args:
            fileobj: the file object to write to. Must be open for writing AND
                reading, and must be positioned at the writing position. Rather
                than initializing directly from a file object, it is recommended
                to use the from_file() constructor.
            marker: the marker to use in headers.
        """
        self.__f = fileobj
        self.__header = _START_TEMPLATE % marker
        self.__footer = _END_TEMPLATE % marker

    def close(self):
        self.__f.close()

    def __enter__(self):
        return self

    def __exit__(self, unused_type, unused_value, traceback):
        self.close()

    @classmethod
    def from_file(cls, filename, marker, append=False):
        """Construct a writer for writing to a file.

        Caller is responsible for closing the writer afterwards.

        Args:
            filename: the file to write to.
            marker: the marker to use in headers/footers.
            append: if True, file will be opened in append mode.

        Returns:
            A PemWriter object.

        Raises:
            IOError: the file could not be opened.
        """
        mode = "a+" if append else "w+"
        f = open(filename, mode)
        if append:
            f.seek(0, 2)
        return cls(f, marker)

    def write(self, blob, check_newline=True):
        """Write a single PEM blob.

        Args:
            blob: a binary blob.
            check_newline: if True, check whether the current position is at
                the beginning of a new line and add a newline if not.

        Raises:
            IOError: the file could not be written to.
        """
        # Header must start on a new line, so we try to be helpful and add one
        # if it's missing.
        # Note that a file open'ed in a+ mode will report its current reading
        # (rather than writing) position - we deem it the caller's
        # responsibility to seek to the write position. Failing that, the worst
        # that can happen is either we fail to heal it, or add an extra newline.
        if check_newline:
          if self.__f.tell() != 0:
              self.__f.seek(-1, 1)
              if self.__f.read(1) != "\n":
                  self.__f.write("\n")

        self.__f.write(self.__header)
        pem_blob = base64.b64encode(blob)
        for i in range(0, len(pem_blob), 64):
            self.__f.write("\n")
            self.__f.write(pem_blob[i:i+64])

        self.__f.write("\n")
        self.__f.write(self.__footer)
        self.__f.write("\n")

    def write_blocks(self, blobs):
        """Write PEM blobs.

        Args:
            blobs: an iterable of binary blobs.

        Raises:
            IOError: the file could not be written to.
        """
        check_newline = True
        for b in blobs:
            self.write(b, check_newline=check_newline)
            check_newline = False

    @classmethod
    def pem_string(cls, blob, marker):
        """Convert a binary blob to a PEM string.

        Args:
            blob: a single binary blob.
            marker: the marker to use in headers/footers.

        Returns:
            a string of concatenated PEM blobs.
        """
        stringio = StringIO.StringIO()
        with cls(stringio, marker) as writer:
            writer.write(blob)
            return stringio.getvalue()

    @classmethod
    def blocks_to_pem_string(cls, blobs, marker):
        """Convert a binary blob to a PEM string.

        Args:
            blobs: an iterable of binary blobs.
            marker: the marker to use in headers/footers.

        Returns:
            a string of concatenated PEM blobs.
        """
        stringio = StringIO.StringIO()
        with cls(stringio, marker) as writer:
            writer.write_blocks(blobs)
            return stringio.getvalue()


def from_pem(pem_string, markers):
    """Read a single PEM blob from a string.

    Ignores everything before and after the first blob with valid markers.

    Args:
        pem_string: the PEM string.
        markers: a single marker string or an iterable containing all
            accepted markers, such as CERTIFICATE, RSA PUBLIC KEY,
            PUBLIC KEY, etc.

    Returns:
        A (raw_string, marker) tuple containing the decoded blob and the
        marker used to detect the blob.

    Raises:
        PemError: a PEM block was invalid or no valid PEM block was found.
    """
    with PemReader.from_string(pem_string, markers,
                               skip_invalid_blobs=False) as reader:
        return iter(reader).next()


def from_pem_file(pem_file, markers):
    """Read a single PEM blob from a file.

    Ignores everything before and after the first blob with valid markers.

    Args:
        pem_file: the PEM file.
        markers: a single marker string or an iterable containing all
            accepted markers, such as CERTIFICATE, RSA PUBLIC KEY,
            PUBLIC KEY, etc.

    Returns:
        A (raw_string, marker) tuple containing the decoded blob and the
        marker used to detect the blob.

    Raises:
        PemError: a PEM block was invalid or no valid PEM block was found.
        IOError: the file could not be read.
    """
    with PemReader.from_file(pem_file, markers,
                             skip_invalid_blobs=False) as reader:
        return iter(reader).next()


def pem_blocks(pem_string, markers, skip_invalid_blobs=True):
    """Read PEM blobs from a string.

    Args:
        pem_string: the PEM string.
        markers: a single marker string or an iterable containing all
            accepted markers, such as CERTIFICATE, RSA PUBLIC KEY,
            PUBLIC KEY, etc.
        skip_invalid_blobs: if False, invalid PEM blobs cause a PemError.
            If True, invalid blobs are skipped. In non-skip mode, an immediate
            StopIteration before any valid blocks are found, also causes a
            a PemError exception.

    Yields:
        (raw_string, marker) tuples containing the decoded blob and the marker
        used to detect the blob.

    Raises:
        PemError: a PEM block was invalid.
    """
    with PemReader.from_string(pem_string, markers,
                               skip_invalid_blobs=skip_invalid_blobs) as reader:
        for block in reader:
            yield block


def pem_blocks_from_file(pem_file, markers, skip_invalid_blobs=True):
    """Read PEM blobs from a file.

    Args:
        pem_file: the PEM file.
        markers: a single marker string or an iterable containing all accepted
            markers, such as CERTIFICATE, RSA PUBLIC KEY, PUBLIC KEY, etc.
        skip_invalid_blobs: if False, invalid PEM blobs cause a PemError.
            If True, invalid blobs are skipped. In non-skip mode, an immediate
            StopIteration before any valid blocks are found, also causes a
            PemError exception.

    Yields:
        (raw_string, marker) tuples containing the decoded blob and the marker
        used to detect the blob.

    Raises:
        PemError: a PEM block was invalid.
    """
    with PemReader.from_file(pem_file, markers,
                             skip_invalid_blobs=skip_invalid_blobs) as reader:
        for block in reader:
            yield block


def to_pem(blob, marker):
    """Convert a binary blob to a PEM-formatted string.

    Args:
        blob: a binary blob.
        marker: the marker to use, e.g., CERTIFICATE.

    Returns:
        the PEM string.
    """
    return PemWriter.pem_string(blob, marker)


def blocks_to_pem(blobs, marker):
    """Convert binary blobs to a string of concatenated PEM-formatted blocks.

    Args:
        blobs: an iterable of binary blobs
        marker: the marker to use, e.g., CERTIFICATE

    Returns:
        the PEM string.
    """
    return PemWriter.blocks_to_pem_string(blobs, marker)


def to_pem_file(blob, filename, marker):
    """Convert a binary blob to PEM format and write to file.

    Args:
        blob: a binary blob.
        filename: the file to write to.
        marker: the marker to use, e.g., CERTIFICATE.

    Raises:
        IOError: the file could not be written to.
    """
    with PemWriter.from_file(filename, marker) as writer:
        writer.write(blob)


def blocks_to_pem_file(blobs, filename, marker):
    """Convert binary blobs to PEM format and write to file.

    Blobs must all be of one and the same type.

    Args:
        blobs: an iterable of binary blobs.
        filename: the file to write to.
        marker: the marker to use, e.g., CERTIFICATE.

    Raises:
        IOError: the file could not be written to.
    """
    with PemWriter.from_file(filename, marker) as writer:
        writer.write_blocks(blobs)



"""ASN.1 specification for X509 name types."""

from smime.crypto import error
from smime.crypto.asn1 import oid
from smime.crypto.asn1 import types


class AttributeType(oid.ObjectIdentifier):
    pass


class AttributeValue(types.Any):
    pass


class DirectoryString(types.Choice):
    components = {
        "teletexString": types.TeletexString,
        "printableString": types.PrintableString,
        "universalString": types.UniversalString,
        "utf8String": types.UTF8String,
        "bmpString": types.BMPString,
        # Does not really belong here.
        "ia5String": types.IA5String
        }


_ATTRIBUTE_DICT = {
    # Note: this mapping does not conform to the RFCs, as some of the OIDs
    # have more restricted values. But real certificates do not conform either,
    # so we try to be lenient and accept all strings that we can recognize.
    oid.ID_AT_NAME: DirectoryString,
    oid.ID_AT_SURNAME: DirectoryString,
    oid.ID_AT_GIVEN_NAME: DirectoryString,
    oid.ID_AT_INITIALS: DirectoryString,
    oid.ID_AT_GENERATION_QUALIFIER: DirectoryString,
    oid.ID_AT_COMMON_NAME: DirectoryString,
    oid.ID_AT_LOCALITY_NAME: DirectoryString,
    oid.ID_AT_STATE_OR_PROVINCE_NAME: DirectoryString,
    oid.ID_AT_ORGANIZATION_NAME: DirectoryString,
    oid.ID_AT_ORGANIZATIONAL_UNIT_NAME: DirectoryString,
    oid.ID_AT_TITLE: DirectoryString,
    oid.ID_AT_DN_QUALIFIER: DirectoryString,  # PrintableString
    oid.ID_AT_COUNTRY_NAME: DirectoryString,  # PrintableString
    oid.ID_AT_SERIAL_NUMBER: DirectoryString,  # PrintableString
    oid.ID_AT_PSEUDONYM: DirectoryString,
    oid.ID_DOMAIN_COMPONENT: DirectoryString,  # IA5String
    oid.ID_EMAIL_ADDRESS: DirectoryString,  # IA5String
    oid.ID_AT_STREET_ADDRESS: DirectoryString,
    oid.ID_AT_DESCRIPTION: DirectoryString,
    oid.ID_AT_BUSINESS_CATEGORY: DirectoryString,
    oid.ID_AT_POSTAL_CODE: DirectoryString,
    oid.ID_AT_POST_OFFICE_BOX: DirectoryString,
    }


class AttributeTypeAndValue(types.Sequence):
    print_labels = False
    print_delimiter = "="
    components = (
        (types.Component("type", AttributeType)),
        (types.Component("value", AttributeValue, defined_by="type",
                         lookup=_ATTRIBUTE_DICT))
        )


class RelativeDistinguishedName(types.SetOf):
    print_labels = False
    print_delimiter = ", "
    component = AttributeTypeAndValue


class RDNSequence(types.SequenceOf):
    print_labels = False
    print_delimiter = "/"
    component = RelativeDistinguishedName
    # See http://tools.ietf.org/html/rfc6125 for context.
    def flatten(self):
        """Get a flat list of AttributeTypeAndValue pairs in an RDNSequence.

        The hierarchical (Relative) information is not used in all contexts,
        so we provide a way of discarding that information and flattening
        the structure.
        """
        return sum([list(rdn) for rdn in self], [])

    def attributes(self, attr_type):
        """Get a flat list of attribute values of the given type.

        Returns:
            a list of attributes.

        Raises:
            error.ASN1Error: corrupt attribute value.
        """
        attrs = self.flatten()
        decoded_values = [attr["value"].decoded_value for attr in attrs
                          if attr["type"] == attr_type]
        if any([val is None for val in decoded_values]):
            raise error.ASN1Error("Corrupt name attribute")
        # A subject name attribute is always a DirectoryString (a Choice),
        # so we need to take its value.
        return [d.component_value() for d in decoded_values]


# Bypass the CHOICE indirection since exactly one option is specified.
# class Name(types.Choice):
#     components = {"rdnSequence": RDNSequence}
class Name(RDNSequence):
    pass


class OtherName(types.Sequence):
    print_delimiter = ", "
    components = (
        (types.Component("type-id", oid.ObjectIdentifier)),
        (types.Component("value", types.Any.explicit(0)))
        )


class EDIPartyName(types.Sequence):
    print_delimiter = ", "
    components = (
        # Definition here: http://tools.ietf.org/html/rfc5280#section-4.2.1.6
        # Note: this definition suggests that the tagging is implicit.
        # However, implicit tagging of a CHOICE type is ambiguous, so this is
        # in practice interpreted as an explicit tag.
        (types.Component("nameAssigner", DirectoryString.explicit(0),
                         optional=True)),
        (types.Component("partyName", DirectoryString.explicit(1)))
        )


# Partially defined ORAddress: we've not come across any certs that contain it
# but this should be enough to allow the decoder to continue without blowing up.
class BuiltInDomainDefinedAttributes(types.SequenceOf):
    component = types.Any


class ExtensionAttributes(types.SetOf):
    component = types.Any


class ORAddress(types.Sequence):
    components = (
        (types.Component("builtInStandardAttributes", types.Any)),
        (types.Component("builtInDomainDefinedAttributes",
                         BuiltInDomainDefinedAttributes, optional=True)),
        (types.Component("extensionAttributes",
                         ExtensionAttributes, optional=True))
        )


OTHER_NAME = "otherName"
RFC822_NAME = "rfc822Name"
DNS_NAME = "dNSName"
X400_ADDRESS_NAME = "x400Address"
DIRECTORY_NAME = "directoryName"
EDI_PARTY_NAME = "ediPartyName"
URI_NAME = "uniformResourceIdentifier"
IP_ADDRESS_NAME = "iPAddress"
REGISTERED_ID_NAME = "registeredID"


class IPAddress(types.OctetString):
    def __init__(self, value=None, serialized_value=None, strict=True):
        super(IPAddress, self).__init__(value=value,
                                        serialized_value=serialized_value,
                                        strict=strict)
        if strict and len(self._value) != 4 and len(self._value) != 16:
            raise error.ASN1Error("%s is not a valid IP address" %
                                  self.value.encode("hex"))

    def as_octets(self):
        return tuple([ord(b) for b  in self._value])

    def __str__(self):
        if len(self._value) == 4:
            return ".".join([str(ord(c)) for c in self._value])
        if len(self._value) == 16:
            return ":".join([self._value[i:i+2].encode("hex")
                             for i in range(0, len(self._value), 2)])
        return super(IPAddress, self).__str__()


class GeneralName(types.Choice):
    print_labels = True  # Print component type.
    # Definition here: http://tools.ietf.org/html/rfc5280#section-4.2.1.6
    components = {
        OTHER_NAME: OtherName.implicit(0),
        RFC822_NAME: types.IA5String.implicit(1),
        DNS_NAME: types.IA5String.implicit(2),
        X400_ADDRESS_NAME: ORAddress.implicit(3),
        # Implicit CHOICE tag is converted to an explicit one.
        DIRECTORY_NAME: Name.explicit(4),
        EDI_PARTY_NAME: EDIPartyName.implicit(5),
        URI_NAME: types.IA5String.implicit(6),
        IP_ADDRESS_NAME: IPAddress.implicit(7),
        REGISTERED_ID_NAME: oid.ObjectIdentifier.implicit(8)
        }


class GeneralNames(types.SequenceOf):
  component = GeneralName



"""ASN.1 types.

Spec: http://www.itu.int/ITU-T/studygroups/com17/languages/X.690-0207.pdf
See also http://luca.ntop.org/Teaching/Appunti/asn1.html for a good introduction
to ASN.1.

This module implements a restricted encoder/decoder for a subset of ASN.1 types.

The decoder has a strict and non-strict mode. Non-strict mode tolerates selected
non-fatal DER decoding errors. The encoder is DER-only.

Generic decoding is not supported: objects can only be decoded against a
predefined ASN.1 type. However, applications can derive arbitrary custom ASN.1
type specifications from the supported base types.

Constraints (e.g., on the length of an ASN.1 string value) are not supported,
and should be checked at application level, where necessary.
"""

import abc
import collections
import functools
import re

from smime.crypto import error
from smime.crypto.asn1 import print_util
from smime.crypto.asn1 import tag

_ZERO = "\x00"
_MINUS_ONE = "\xff"
_EOC = "\x00\x00"


def encode_int(value, signed=True):
    """Encode an integer.

    Args:
        value: an integral value.
        signed: if True, encode in two's complement form. If False, encode as
            an unsigned integer.

    Raises:
        ValueError: attempt to encode a negative integer as unsigned.

    Returns:
        a variable-length string representing the encoded integer.
    """
    if not signed and value < 0:
        raise ValueError("Unsigned integer cannot be negative")

    if not value:
        return _ZERO
    if value == -1:
        return _MINUS_ONE

    int_bytes = bytearray()

    while value != 0 and value != -1:
        int_bytes.append(value & 0xff)
        value >>= 8

    if signed:
        # In two's complement form, negative values have the most significant
        # bit set, thus we:
        if value == -1 and int_bytes[-1] <= 127:
            # Add a "-1"-byte for indicating a negative value.
            int_bytes.append(0xff)
        elif value == 0 and int_bytes[-1] > 127:
            # Add a "0"-byte for indicating a positive value.
            int_bytes.append(0)

    int_bytes.reverse()
    return str(int_bytes)


def decode_int(buf, signed=True, strict=True):
    """Decode an integer.

    Args:
        buf: a string or string buffer.
        signed: if True, decode in two's complement form. If False, decode as
            an unsigned integer.

    Raises:
        ASN1Error.

    Returns:
        an integer.
    """
    if not buf:
        raise error.ASN1Error("Invalid integer encoding: empty value")

    leading = ord(buf[0])
    int_bytes = bytearray(buf[1:])

    if int_bytes:
        if strict and leading == 0 and int_bytes[0] < 128:
            # 0x00 0x42 == 0x42
            raise error.ASN1Error("Extra leading 0-bytes in integer "
                                  "encoding")
        elif strict and signed and leading == 0xff and int_bytes[0] >= 128:
            # 0xff 0x82 == 0x82
            raise error.ASN1Error("Extra leading 0xff-bytes in negative "
                                  "integer encoding")

    if signed and leading > 127:
        leading -= 256

    for b in int_bytes:
        leading <<= 8
        leading += b

    return leading


# Lengths between 0 and 127 are encoded as a single byte.
# Lengths greater than 127 are encoded as follows:
#   * MSB of first byte is 1 and remaining bits encode the number of
#     additional bytes.
#   * Remaining bytes encode the length.
_MULTIBYTE_LENGTH = 0x80
_MULTIBYTE_LENGTH_MASK = 0x7f


def encode_length(length):
    """Encode an integer.

    Args:
        length: a non-negative integral value.

    Returns:
        a string.
    """
    if length <= 127:
        return chr(length)
    encoded_length = encode_int(length, signed=False)
    return chr(_MULTIBYTE_LENGTH | len(encoded_length)) + encoded_length


def read_length(buf, strict=True):
    """Read an ASN.1 object length from the beginning of the buffer.

    Args:
        buf: a string or string buffer.
        strict: if false, accept indefinite length encoding.

    Raises:
        ASN1Error.

    Returns:
        a (length, rest) tuple consisting of a non-negative integer representing
        the length of an ASN.1 object, and the remaining bytes. For indefinite
        length, returns (-1, rest).
    """
    if not buf:
        raise error.ASN1Error("Invalid length encoding: empty value")
    length, rest = ord(buf[0]), buf[1:]
    if length <= 127:
        return length, rest
    # 0x80 == ASN.1 indefinite length
    if length == 128:
        if strict:
            raise error.ASN1Error("Indefinite length encoding")
        return -1, rest

    length &= _MULTIBYTE_LENGTH_MASK
    if len(rest) < length:
        raise error.ASN1Error("Invalid length encoding")
    # strict=True: let's hope that at least ASN.1 lengths are properly encoded.
    return decode_int(rest[:length], signed=False, strict=True), rest[length:]


class Universal(object):
    """Apply a universal tag to the class.

    Can be used as a callable, or a decorator:

    Integer = Universal(2, tag.PRIMITIVE)(Abstract)

    is the same as

    @Universal(2, tag.PRIMITIVE)
    class Integer(Abstract):
        pass

    and defines a type with an ASN.1 integer tag.
    """

    def __init__(self, number, encoding):
        """Setup the tag.

        Args:
            number: the tag number.
            encoding: the encoding. One of tag.PRIMITIVE or tag.CONSTRUCTED.
        """
        self.tag = tag.Tag(number, tag.UNIVERSAL, encoding)

    def __call__(self, cls):
        """Apply the universal tag.

        Args:
            cls: class to modify. The class must have an empty 'tags'
                attribute.

        Returns:
            the class with a modified 'tags' attribute.

        Raises:
            TypeError: invalid application of the tag.
        """
        if cls.tags:
            raise TypeError("Cannot apply a UNIVERSAL tag to a tagged type.")
        cls.tags = (self.tag,)
        return cls


class Explicit(object):
    """Apply an explicit tag to the class.

    Can be used as a callable, or a decorator:

    MyInteger = Explicit(0, tag.APPLICATION)(Integer)

    is the same as

    @Explicit(0, tag.APPLICATION)
    class MyInteger(Integer):
        pass

    and results in a MyInteger type that is explicitly tagged with an
    application-class 0-tag.
    """

    def __init__(self, number, tag_class=tag.CONTEXT_SPECIFIC):
        """Setup the tag.

        Args:
            number: the tag number.
            tag_class: the tag class. One of tag.CONTEXT_SPECIFIC,
                tag.APPLICATION or tag.PRIVATE.

        Raises:
            TypeError: invalid application of the tag.
        """
        if tag_class == tag.UNIVERSAL:
            raise TypeError("Cannot tag with a UNIVERSAL tag")
        # Explicit tagging always results in constructed encoding.
        self._tag = tag.Tag(number, tag_class, tag.CONSTRUCTED)

    def __call__(self, cls):
        """Apply the explicit tag.

        Args:
            cls: class to modify. The class must have an iterable 'tags'
                attribute.
        Returns:
            the class with a modified 'tags' attribute.
        """
        tags = list(cls.tags)
        tags.append(self._tag)
        cls.tags = tuple(tags)
        return cls


class Implicit(object):
    """Apply an implicit tag to the class.

    Can be used as a callable, or a decorator:

    MyInteger = Implicit(0, tag.APPLICATION)(Integer)

    is the same as

    @Implicit(0, tag.APPLICATION)
    class MyInteger(Integer):
        pass

    and results in a MyInteger type whose tag is implicitly replaced with an
    application-class 0-tag.
    """

    def __init__(self, number, tag_class=tag.CONTEXT_SPECIFIC):
        """Setup the tag.

        Args:
            number: the tag number.
            tag_class: the tag class. One of tag.CONTEXT_SPECIFIC,
                tag.APPLICATION or tag.PRIVATE.

        Raises:
            TypeError: invalid application of the tag.
        """
        if tag_class == tag.UNIVERSAL:
            raise TypeError("Cannot tag with a UNIVERSAL tag")
        # We cannot precompute the tag because the encoding depends
        # on the existing tags.
        self._number = number
        self._tag_class = tag_class

    def __call__(self, cls):
        """Apply the implicit tag.

        Args:
            cls: class to modify. The class must have an iterable 'tags'
                attribute.

        Returns:
            the class with a modified 'tags' attribute.

        Raises:
            TypeError: invalid application of the tag.
        """
        if not cls.tags:
            raise TypeError("Cannot implicitly tag an untagged type")
        tags = list(cls.tags)
        # Only simple types and simple types derived via implicit tagging have a
        # primitive encoding, so the last tag determines the encoding type.
        tags[-1] = (tag.Tag(self._number, self._tag_class,
                            cls.tags[-1].encoding))
        cls.tags = tuple(tags)
        return cls


class Abstract(object):
    """Abstract base class."""
    __metaclass__ = abc.ABCMeta

    tags = ()

    @classmethod
    def explicit(cls, number, tag_class=tag.CONTEXT_SPECIFIC):
        """Dynamically create a new tagged type.

        Args:
            number: tag number.
            tag_class: tag class.

        Returns:
            a subtype of cls with the given explicit tag.
        """
        name = "%s.explicit(%d, %d)" % (cls.__name__, number, tag_class)

        # TODO(ekasper): the metaclass could register created types so we
        # return the _same_ type when called more than once with the same
        # arguments.
        mcs = cls.__metaclass__
        return_class = mcs(name, (cls,), {})
        return Explicit(number, tag_class)(return_class)

    @classmethod
    def implicit(cls, number, tag_class=tag.CONTEXT_SPECIFIC):
        """Dynamically create a new tagged type.

        Args:
            number: tag number.
            tag_class: tag class.

        Returns:
            a subtype of cls with the given implicit tag.
        """
        name = "%s.implicit(%d, %d)" % (cls.__name__, number, tag_class)
        mcs = cls.__metaclass__
        return_class = mcs(name, (cls,), {})
        return Implicit(number, tag_class)(return_class)

    def __init__(self, value=None, serialized_value=None, strict=True):
        """Initialize from a value or serialized buffer.

        Args:
            value: initializing value of an appropriate type. If the
                serialized_value is not set, the initializing value must be set.
            serialized_value: serialized inner value (with tags and lengths
                stripped).
            strict: if False, tolerate some non-fatal decoding errors.

        Raises:
            error.ASN1Error: decoding the serialized value failed.
            TypeError: invalid initializer.
        """
        if serialized_value is not None:
            self._value = self._decode_value(serialized_value, strict=strict)
        elif value is not None:
            self._value = self._convert_value(value)
        else:
            raise TypeError("Cannot initialize from None")
        self._serialized_value = serialized_value

    @classmethod
    def _convert_value(cls, value):
        """Convert initializer to an appropriate value."""
        raise NotImplementedError

    @abc.abstractmethod
    def _decode_value(self, buf, strict=True):
        """Decode the initializer value from a buffer.

        Args:
            buf: a string or string buffer.
            strict: if False, tolerate some non-fatal decoding errors.

        Returns:
           the value of the object.
        """
        pass

    @property
    def value(self):
        """Get the value of the object.

        An ASN.1 object can always be reconstructed from its value.
        """
        # Usually either the immutable value, or a shallow copy of
        # the mutable value.
        raise NotImplementedError

    def __repr__(self):
        return "%s(%r)" % (self.__class__.__name__, self.value)

    def __str__(self):
        return str(self.value)

    @abc.abstractmethod
    def _encode_value(self):
        """Encode the contents, excluding length and tags.

        Returns:
            a string representing the encoded value.
        """
        pass

    # Implemented by Choice and Any.
    # Used when the type is untagged so that read() does not reveal a length.
    @classmethod
    def _read(cls, buf, strict=True):
        """Read the value from the beginning of a string or buffer."""
        raise NotImplementedError

    # Only applicable where indefinite length encoding is possible, i.e., for
    # simple string types using constructed encoding and structured types
    # (Sequence, Set, SequenceOf, SetOf) only. Currently, it's only
    # implemented for structured types; constructed encoding of simple string
    # types is not supported. Only applicable in non-strict mode.
    @classmethod
    def _read_indefinite_value(cls, buf):
        """Read the inner value from the beginning of a string or buffer."""
        raise NotImplementedError

    def encode(self):
        """Encode oneself.

        Returns:
            a string representing the encoded object.
        """
        # If we have a read-only object that we created from a serialized value
        # and never modified since, use the original cached value.
        #
        # This ensures that objects decoded in non-strict mode will retain their
        # original encoding.
        #
        # BUG: we do not cache tag and length encoding, so reencoding is broken
        # for objects that use indefinite length encoding.
        if self._serialized_value and not self.modified():
            encoded_value = self._serialized_value
        else:
            # We can only use the cached value if the object has never been
            # modified after birth. Since mutable objects cannot track when
            # their recursive subcomponents are modified, the modified flag,
            # once set, can never be unset.
            self._serialized_value = None
            encoded_value = self._encode_value()
        for t in self.tags:
            encoded_length = encode_length(len(encoded_value))
            encoded_value = t.value + encoded_length + encoded_value
        return encoded_value

    @classmethod
    def read(cls, buf, strict=True):
        """Read from a string or buffer.

        Args:
            buf: a string or string buffer.
            strict: if False, tolerate some non-fatal decoding errors.

        Returns:
            a tuple consisting of an instance of the class and the remaining
            bytes.
        """
        if cls.tags:
            # Each indefinite length must be closed with the EOC (\x00\x00)
            # octet.
            # If we have multiple tags (i.e., explicit tagging is used) and the
            # outer tags use indefinite length, each such encoding adds an EOC
            # to the end (while a regular tag adds nothing). Therefore, we first
            # read all tags, then the value, and finally strip the EOC octets of
            # the explicit tags.
            indefinite = 0
            decoded_length = -1
            for t in reversed(cls.tags):
                if buf[:len(t)] != t.value:
                    raise error.ASN1TagError(
                        "Invalid tag: expected %s, got %s while decoding %s" %
                        (t, buf[:len(t.value)], cls.__name__))
                # Logging statements are really expensive in the recursion even
                # if debug-level logging itself is disabled.
                # logging.debug("%s: read tag %s", cls.__name__, t)
                buf = buf[len(t):]
                # Only permit indefinite length for constructed types.
                decoded_length, buf = read_length(buf, strict=(
                    strict or t.encoding != tag.CONSTRUCTED))
                if decoded_length == -1:
                    indefinite += 1
                # logging.debug("%s: read length %d", cls.__name__,
                #               decoded_length)
                elif len(buf) < decoded_length:
                    raise error.ASN1Error("Invalid length encoding in %s: "
                                          "read length %d, remaining bytes %d" %
                                          (cls.__name__, decoded_length,
                                           len(buf)))

            # The last tag had definite length.
            if decoded_length != -1:
                value, rest = (cls(serialized_value=buf[:decoded_length],
                                   strict=strict), buf[decoded_length:])
            else:
                decoded, rest = cls._read_indefinite_value(buf)
                value = cls(value=decoded)
                # _read_indefinite_value will strip the inner EOC.
                indefinite -= 1
            # Remove EOC octets corresponding to outer explicit tags.
            if indefinite:
                if rest[:indefinite * 2] != _EOC * indefinite:
                    raise error.ASN1Error("Missing EOC octets")
                rest = rest[indefinite * 2:]

        else:
            # Untagged CHOICE and ANY; no outer tags to determine the length.
            value, rest = cls._read(buf, strict=strict)

        # logging.debug("%s: decoded value %s", cls.__name__, value)
        # logging.debug("Remaining bytes: %d", len(rest))
        return value, rest

    @classmethod
    def decode(cls, buf, strict=True):
        """Decode from a string or buffer.

        Args:
            buf: a string or string buffer.
            strict: if False, tolerate some non-fatal decoding errors.

        Returns:
            an instance of the class.
        """
        value, rest = cls.read(buf, strict=strict)
        if rest:
            raise error.ASN1Error("Invalid encoding: leftover bytes when "
                                  "decoding %s" % cls.__name__)
        return value

    # Compare by value.
    # Note this means objects with equal values do not necessarily have
    # equal encodings.
    def __eq__(self, other):
        return self.value == other

    def __ne__(self, other):
        return self.value != other

    @abc.abstractmethod
    def human_readable_lines(self, wrap=80, label=""):
        """A pretty human readable representation of the object.

        Args:
            wrap: maximum number of characters per line. 0 or negative wrap
                means no limit. Should be chosen long enough to comfortably fit
                formatted data; otherwise it is simply ignored and output may
                look funny.
            label: a label prefix.

        Returns:
            a list of line strings of at most |wrap| characters each.
        """
        return []

    def human_readable(self, wrap=80, label=""):
        """A pretty human readable representation of the object.

        Args:
            wrap: maximum number of characters per line. 0 or negative wrap
               means no limit. Should be chosen long enough to comfortably fit
               formatted data; otherwise it is simply ignored and output may
               look funny.
            label: a label prefix.

        Returns:
            a multi-line string of at most |wrap| characters per line.
        """
        return ("\n").join(self.human_readable_lines(wrap=wrap, label=label))

    @abc.abstractmethod
    def modified(self):
        """Returns True if the object has been modified after creation."""
        pass


# Boilerplate code for some simple types whose value directly corresponds to a
# basic immutable type.
@functools.total_ordering
class Simple(Abstract):
    """Base class for Boolean, Integer, and string types."""
    # Pretty-printed character length.
    # OctetString and BitString use this to nicely format hex bytes.
    char_wrap = 1

    @property
    def value(self):
        return self._value

    def __hash__(self):
        return hash(self.value)

    def __lt__(self, other):
        return self.value < other

    def __bool__(self):
        return bool(self.value)

    def __int__(self):
        return int(self.value)

    def __nonzero__(self):
        return bool(self.value)

    def modified(self):
        """Returns True if the object has been modified after creation."""
        return False

    @classmethod
    def wrap_lines(cls, long_string, wrap):
        """Split long lines into multiple chunks according to the wrap limit.

        Derived classes can override char_wrap if they wish to, e.g., not split
        hex bytes.

        Args:
            long_string: a string_value() representation of the object
            wrap: maximum number of characters per line. 0 or negative wrap
                means no limit. Should be chosen long enough to comfortably fit
                formatted data; otherwise it is simply ignored and output may
                look funny.

        Returns:
           long_string split into lines of at most |wrap| characters each.
        """
        wrap -= wrap % cls.char_wrap
        return print_util.wrap_lines(long_string, wrap)

    def human_readable_lines(self, wrap=80, label=""):
        """A pretty human readable representation of the object.

        Args:
            wrap: maximum number of characters per line. 0 or negative wrap
                means no limit. Should be chosen long enough to comfortably fit
                formatted data; otherwise it is simply ignored and output may
                look funny.
            label: a label prefix.

        Returns:
            a list of line strings of at most |wrap| characters each.
        """
        to_print = str(self)
        formatted_label = label + ": " if label else ""
        if (to_print.find("\n") == -1 and
                (wrap <= 0 or len(to_print) + len(formatted_label) <= wrap)):
            # Fits on one line, like this:
            # label: value
            return [formatted_label + to_print]

        else:
            # Multiline output:
            # label:
            #   firstlongvalueline
            #   secondvalueline
            ret = []
            indent = 2
            if label:
                ret += print_util.wrap_lines(label + ":", wrap)
            return ret + [" " * indent + x for x in
                          self.wrap_lines(to_print, wrap - indent)]


@Universal(1, tag.PRIMITIVE)
class Boolean(Simple):
    """Boolean."""
    _TRUE = "\xff"
    _FALSE = "\x00"

    def _encode_value(self):
        return self._TRUE if self._value else self._FALSE

    @classmethod
    def _convert_value(cls, value):
        return bool(value)

    @classmethod
    def _decode_value(cls, buf, strict=True):
        if len(buf) != 1:
            raise error.ASN1Error("Invalid encoding")

        # Continuing here breaks re-encoding.
        if strict and buf[0] != cls._TRUE and buf[0] != cls._FALSE:
            raise error.ASN1Error("BER encoding of Boolean value: %s" %
                                  buf[0])
        value = False if buf[0] == cls._FALSE else True
        return value


@Universal(2, tag.PRIMITIVE)
class Integer(Simple):
    """Integer."""

    def _encode_value(self):
        return encode_int(self._value)

    @classmethod
    def _convert_value(cls, value):
        return int(value)

    @classmethod
    def _decode_value(cls, buf, strict=True):
        return decode_int(buf, strict=strict)


@Universal(5, tag.PRIMITIVE)
class Null(Simple):
    """Null."""

    def _encode_value(self):
        return ""

    @classmethod
    def _convert_value(cls, value):
        return None

    @classmethod
    def _decode_value(cls, buf, strict=True):
        return None


class ASN1String(Simple):
    """Base class for string types."""

    def _encode_value(self):
        return self._value

    @classmethod
    def _convert_value(cls, value):
        if isinstance(value, str) or isinstance(value, buffer):
            value = str(value)
        elif isinstance(value, ASN1String):
            value = value.value
        else:
            raise TypeError("Cannot convert %s to %s" %
                            (type(value), cls.__name__))
        cls._check_for_illegal_characters(value)
        return value

    @classmethod
    def _check_for_illegal_characters(cls, buf):
        """Raises if there are any illegal characters in string.

        Args:
            buf: string which will be checked for illegal characters

        Raises:
            ASN1Error.
        """
        pass

    @classmethod
    def _decode_value(cls, buf, strict=True):
        if strict:
            cls._check_for_illegal_characters(buf)
        return buf


# Based on https://www.itu.int/rec/T-REC-X.208-198811-W/en
# and http://kikaku.itscj.ipsj.or.jp/ISO-IR/overview.htm
@Universal(19, tag.PRIMITIVE)
class PrintableString(ASN1String):
    """PrintableString."""
    NOT_ACCEPTABLE = re.compile("[^a-zA-Z0-9 '()+,\-./:=?]")

    @classmethod
    def _check_for_illegal_characters(cls, buf):
        search_result = PrintableString.NOT_ACCEPTABLE.search(buf)
        if search_result:
            index = search_result.start()
            raise error.ASN1IllegalCharacter(
                "Illegal character in PrintableString", buf, index)


@Universal(20, tag.PRIMITIVE)
class TeletexString(ASN1String):
    """TeletexString (aka T61String)."""
    pass


@Universal(22, tag.PRIMITIVE)
class IA5String(ASN1String):
    """IA5String."""

    @classmethod
    def _check_for_illegal_characters(self, buf):
        for index, character in enumerate(buf):
            if ord(character) > 127:
                raise error.ASN1IllegalCharacter(
                    "Illegal character in IA5String", buf, index)


@Universal(26, tag.PRIMITIVE)
class VisibleString(ASN1String):
    """VisibleString (aka ISO646String)."""

    @classmethod
    def _check_for_illegal_characters(self, buf):
        for index, character in enumerate(buf):
            if ord(character) < 32 or ord(character) > 126:
                raise error.ASN1IllegalCharacter(
                    "Illegal character in VisibleString", buf, index)


@Universal(30, tag.PRIMITIVE)
class BMPString(ASN1String):
    """BMPString."""
    pass


@Universal(12, tag.PRIMITIVE)
class UTF8String(ASN1String):
    """UTF8String."""
    pass


@Universal(28, tag.PRIMITIVE)
class UniversalString(ASN1String):
    """UniversalString."""
    pass


@Universal(4, tag.PRIMITIVE)
class OctetString(ASN1String):
    """Octet string."""
    char_wrap = 3

    def __str__(self):
        return print_util.bytes_to_hex(self._value)


@Universal(3, tag.PRIMITIVE)
class BitString(Simple):
    """Bit string."""
    char_wrap = 3

    def __str__(self):
        return print_util.bits_to_hex(self._value)

    def _encode_value(self):
        pad = (8 - len(self._value) % 8) % 8
        padded_bits = self._value + pad * "0"
        ret = bytearray([pad])
        for i in range(0, len(padded_bits), 8):
            ret.append(int(padded_bits[i:i + 8], 2))
        return str(ret)

    def _convert_value(self, value):
        """The value of a BitString is a string of '0's and '1's."""
        if isinstance(value, BitString):
            return value.value
        elif isinstance(value, str):
            # Must be a string of '0's and '1's.
            if not all(c == "0" or c == "1" for c in value):
                raise ValueError("Cannot initialize a BitString from %s:"
                                 "string must consist of 0s and 1s" % value)
            return value
        else:
            raise TypeError("Cannot initialize a BitString from %s"
                            % type(value))

    @classmethod
    def _decode_value(cls, buf, strict=True):
        if not buf:
            raise error.ASN1Error("Invalid encoding: empty %s value" %
                                  cls.__name__)
        int_bytes = bytearray(buf)
        pad = int_bytes[0]
        if pad > 7:
            raise error.ASN1Error("Invalid padding %d in %s" %
                                  (pad, cls.__name__))
        ret = "".join(format(b, "08b") for b in int_bytes[1:])
        if pad:
            if not ret or any([c == "1" for c in ret[-1 * pad:]]):
                raise error.ASN1Error("Invalid padding")
            ret = ret[:-1 * pad]
        return ret


class NamedBitList(BitString):
    """A bit string with named bits."""
    # To use the NamedBitList ASN.1 construct, set named_bit_list
    # to a tuple of NamedValue instances, where the name of each NamedValue
    # corresponds to the identifier and the value to the number of the
    # distinguished bit, defined by "number" or "DefinedValue" in ASN.1,
    # see http://www.itu.int/ITU-T/studygroups/com17/languages/X.680-0207.pdf
    named_bit_list = None
    char_wrap = 1

    def __str__(self):
        return ", ".join(["%s" % n.name for n in self.bits_set()])

    def has_bit_set(self, number):
        """Test if the given bit is set.

        Args:
            number: the number of the ASN.1 bit. Bit numbering follows ASN.1
                conventions, i.e., bit number 0 is the "leading bit".

        Returns:
            True: the bit is 1.
            False: the bit is 0, or the BitString is not long enough.
        """
        # According to
        # http://www.itu.int/ITU-T/studygroups/com17/languages/X.680-0207.pdf
        # we must not assume that the presence of named bits constrains the
        # contents of the bit string:
        # "21.6 The presence of a "NamedBitList" has no effect on the set of
        # abstract values of this type. Values containing 1 bits other than the
        # named bits are permitted.
        # 21.7 When a "NamedBitList" is used in defining a bitstring type ASN.1
        # encoding rules are free to add (or remove) arbitrarily any trailing 0
        # bits to (or from) values that are being encoded or decoded.
        # Application designers should therefore ensure that different semantics
        # are not associated with such values which differ only in the number of
        # trailing 0 bits.
        return len(self._value) > number and self._value[number] == "1"

    def bits_set(self):
        """List the named_bit_list elements whose bit is set."""
        return [n for n in self.named_bit_list if self.has_bit_set(n.value)]


class Any(ASN1String):
    """Any.

    Any is a container for an arbitrary value. An Any type can be tagged with
    explicit tags like any other type: those tags will be applied to the
    underlying value. Implicit tagging of Any types is not supported.

    The value of an Any is an undecoded raw string. In addition, Any can hold
    the decoded value of the object.
    """
    char_wrap = 3

    def __init__(self, value=None, serialized_value=None, strict=True):
        if isinstance(value, str):
            super(Any, self).__init__(value=None, serialized_value=value,
                                      strict=strict)
            self._decoded_value = None
        else:
            super(Any, self).__init__(value=value,
                                      serialized_value=serialized_value,
                                      strict=strict)
            self._decoded_value = value

    def __repr__(self):
        if self._decoded_value is not None:
            return "%s(%r)" % (self.__class__.__name__, self._decoded_value)
        return "%s(%r)" % (self.__class__.__name__, self._value)

    def __str__(self):
        if self._decoded_value is not None:
            return str(self._decoded_value)
        return print_util.bytes_to_hex(self._value)

    def human_readable_lines(self, wrap=80, label=""):
        """A pretty human readable representation of the object.

        Args:
            wrap: maximum number of characters per line. 0 or negative wrap
                means no limit. Should be chosen long enough to comfortably fit
                formatted data; otherwise it is simply ignored and output may
                look funny.
            label: a label prefix.

        Returns:
            a list of line strings of at most |wrap| characters each.
        """
        if self._decoded_value is not None:
            return self._decoded_value.human_readable_lines(wrap=wrap,
                                                            label=label)
        return super(Any, self).human_readable_lines(wrap=wrap, label=label)

    def modified(self):
        if self._decoded_value is not None:
            return self._decoded_value.modified()
        return False

    def _encode_value(self):
        if self._decoded_value is not None and self._decoded_value.modified():
            return self._decoded_value.encode()
        return self._value

    @property
    def decoded(self):
        return self._decoded_value is not None

    @property
    def decoded_value(self):
        return self._decoded_value

    @classmethod
    def _read(cls, buf, strict=True):
        readahead_tag, rest = tag.Tag.read(buf)
        length, rest = read_length(rest, strict=(
            strict or readahead_tag.encoding != tag.CONSTRUCTED))
        if length == -1:
            # Not sure if this even makes any sense.
            raise NotImplementedError("Indefinite length encoding of ANY types "
                                      "is not supported")
        if len(rest) < length:
            raise error.ASN1Error("Invalid length encoding")
        decoded_length = len(buf) - len(rest) + length
        return cls(serialized_value=buf[:decoded_length],
                   strict=strict), buf[decoded_length:]

    @classmethod
    def _convert_value(cls, value):
        """The value of an Any is the undecoded value."""
        # Always return the undecoded value for consistency; the
        # decoded/decoded_value properties can be used to retrieve the
        # decoded contents.
        if isinstance(value, Any):
            # This gets ambiguous real fast (do we keep the original tags or
            # replace with our own tags?) so we ban it.
            raise TypeError("Instantiating Any from another Any is illegal")
        elif isinstance(value, Abstract):
            return value.encode()
        else:
            raise TypeError("Cannot convert %s to %s" % (type(value),
                                                         cls.__name__))

    @classmethod
    def _decode_value(cls, buf, strict=True):
        return buf

    def decode_inner(self, value_type, strict=True):
        """Decode the undecoded contents according to a given specification.

        Args:
            value_type: an ASN.1 type.
            strict: if False, tolerate some non-fatal decoding errors.

        Raises:
            ASN1Error: decoding failed.
            RuntimeError: value already decoded.
        """
        self._decoded_value = value_type.decode(self._value, strict=strict)


class Constructed(Abstract):
    """Constructed types."""
    print_labels = True
    print_delimiter = "\n"

    def __init__(self, value=None, serialized_value=None, strict=True):
        """Initialize from a value or serialized buffer.

        Args:
            value: initializing value of an appropriate type. If the
                serialized_value is not set, the initializing value must be set.
            serialized_value: serialized inner value (with tags and lengths
                stripped).
            strict: if False, tolerate some non-fatal decoding errors.

        Raises:
            error.ASN1Error: decoding the serialized value failed.
            TypeError: invalid initializer.
        """
        super(Constructed, self).__init__(value=value,
                                          serialized_value=serialized_value,
                                          strict=strict)
        # All methods that mutate the object must set this to True.
        self._modified = False

    def modified(self):
        return self._modified or any([v and v.modified()
                                      for _, v in self.items()])

    def human_readable_lines(self, wrap=80, label=""):
        """A pretty human readable representation of the object.

        Args:
            wrap: maximum number of characters per line. 0 or negative wrap
                means no limit. Should be chosen long enough to comfortably fit
                formatted data; otherwise it is simply ignored and output may
                look funny.
            label: a label prefix.

        Returns:
            a list of line strings of at most |wrap| characters each.
        """
        # A "\n" becomes ["", ""] which magically starts a new line when we call
        # append_lines() on it. Things like "\n-----\n" work, too.
        delimiter = (print_util.wrap_lines(self.print_delimiter, wrap=wrap))
        lines = []

        # Component count. Needed so we can print "<no components>" when none
        # are found.
        count = 0
        # Whether the next component should start on a new line. Set to true
        # when the previous component was multiline. For example, a mix of short
        # and long components with a ", " delimiter is thus printed as
        # short1, short2, short3,
        # myextremelylongcomponentth
        # atspansmultiplelines
        # short4, short5
        newline = False

        if label:
            lines += print_util.wrap_lines(label + ":", wrap)
            # If the delimiter is multiline, then output looks prettier if the
            # label is also on a separate line.
            if len(delimiter) > 1:
                newline = True
            elif len(lines[-1]) < wrap:
                # Else add a whitespace so we get "label: value"
                lines[-1] += " "

        indent = 2
        for key, value in self.items():
            if value is None:
                continue
            label = str(key) if self.print_labels else ""
            print_component = value.human_readable_lines(wrap=wrap - indent,
                                                         label=label)
            if not print_component:
                continue

            if count:
                print_util.append_lines(delimiter, wrap, lines)
            count += 1
            # Make multiline components a separate block on a new line, unless
            # we already are on a new line.
            if (newline or len(print_component) > 1) and lines and lines[-1]:
                lines += print_component
            else:
                print_util.append_lines(print_component, wrap, lines)

            newline = len(print_component) > 1

        if not count:
            print_util.append_lines(["<no components>"], wrap, lines)

        # Indent everything apart from the first line.
        return [lines[0]] + ["  " + x for x in lines[1:]]


class MetaChoice(abc.ABCMeta):
    """Metaclass for building a Choice type."""

    def __new__(mcs, name, bases, dic):
        # Build a tag -> component_name map for the decoder.
        components = dic.get("components", {})
        if components:
            tag_map = {}
            keys_seen = set()
            for key, spec in components.items():
                if key in keys_seen:
                    raise TypeError("Duplicate name in Choice specification")
                keys_seen.add(key)

                if not spec.tags:
                    raise TypeError("Choice type cannot have untagged "
                                    "components")
                if spec.tags[-1] in tag_map:
                    raise TypeError("Duplicate outer tag in a Choice "
                                    "specification")
                tag_map[spec.tags[-1]] = key
            dic["tag_map"] = tag_map
        return super(MetaChoice, mcs).__new__(mcs, name, bases, dic)


class Choice(Constructed, collections.MutableMapping):
    """Choice."""
    __metaclass__ = MetaChoice

    # There is only ever one component anyway.
    print_delimiter = ""
    print_labels = False

    def __init__(self, value=None, serialized_value=None,
                 readahead_tag=None, readahead_value=None, strict=True):
        """Initialize fully or partially.

        Args:
            value: if present, should be a dictionary with one entry
                representing the chosen key and value.
            serialized_value: if present, the serialized contents (with tags
                and lengths stripped).
            readahead_tag: if present, the first tag in serialized_value
            readahead_value: if present, the value wrapped by the first tag in
                serialized value.
            strict: if False, tolerate some non-fatal decoding errors.

        Raises:
            ValueError: invalid initializer value.
        """
        if readahead_tag is not None:
            self._value = self._decode_readahead_value(
                serialized_value, readahead_tag, readahead_value,
                strict=strict)
            self._serialized_value = serialized_value
            self._modified = False
        else:
            super(Choice, self).__init__(value=value,
                                         serialized_value=serialized_value,
                                         strict=strict)

    def __getitem__(self, key):
        value = self._value.get(key, None)
        if value is not None:
            return value
        elif key in self.components:
            return None
        raise KeyError("Invalid key %s for %s" % (key, self.__class__.__name__))

    def __setitem__(self, key, value):
        spec = self.components[key]
        if value is None:
            self._value = {}
        elif type(value) is spec:
            self._value = {key: value}
        # If the supplied value is not of the exact same type then we try to
        # construct one.
        else:
            self._value = {key: spec(value)}
        self._modified = True

    def __delitem__(self, key):
        if key in self._value:
            self._value = {}
        # Raise if the key is invalid; else do nothing.
        elif key not in self.components:
            raise KeyError("Invalid key %s" % key)
        self._modified = True

    def __iter__(self):
        return iter(self._value)

    def __len__(self):
        return len(self._value)

    @property
    def value(self):
        return dict(self._value)

    def component_key(self):
        if not self._value:
            return None
        return self._value.keys()[0]

    # A slightly unfortunate overload of the term "value"...
    def component_value(self):
        if not self._value:
            return None
        return self._value.values()[0]

    def _encode_value(self):
        if not self._value:
            raise error.ASN1Error("Choice component not set")
        # Encode the single component.
        return self._value.values()[0].encode()

    @classmethod
    def _read(cls, buf, strict=True):
        readahead_tag, rest = tag.Tag.read(buf)
        length, rest = read_length(rest, strict=(
            strict or readahead_tag.encoding != tag.CONSTRUCTED))
        if length == -1:
            raise NotImplementedError("Indefinite length encoding of CHOICE "
                                      "type is not supported")
        if len(rest) < length:
            raise error.ASN1Error("Invalid length encoding")
        decoded_length = len(buf) - len(rest) + length
        return (cls(serialized_value=buf[:decoded_length],
                    readahead_tag=readahead_tag, readahead_value=rest[:length],
                    strict=strict),
                buf[decoded_length:])

    @classmethod
    def _convert_value(cls, value):
        if not value:
            return dict()
        if len(value) != 1:
            raise ValueError("Choice must have at most one component set")

        key, value = value.items()[0]
        if value is None:
            return {}

        try:
            spec = cls.components[key]
        except KeyError:
            raise ValueError("Invalid Choice key %s" % key)
        if type(value) is spec:
            return {key: value}
        # If the supplied value is not of the exact same type then we try to
        # construct one.
        else:
            return {key: spec(value)}

    @classmethod
    def _decode_readahead_value(cls, buf, readahead_tag, readahead_value,
                                strict=True):
        """Decode using additional information about the outermost tag."""
        try:
            key = cls.tag_map[readahead_tag]
        except KeyError:
            raise error.ASN1TagError("Tag %s is not a valid tag for a "
                                     "component of %s" %
                                     (readahead_tag, cls.__name__))

        if len(cls.components[key].tags) == 1:
            # Shortcut: we already know the tag and length, so directly get
            # the value.
            value = cls.components[key](serialized_value=readahead_value,
                                        strict=strict)
        else:
            # Component has multiple tags but the readahead only read the
            # outermost tag, so read everything again.
            value, rest = cls.components[key].read(buf, strict=strict)
            if rest:
                raise error.ASN1Error("Invalid encoding: leftover bytes when "
                                      "decoding %s" % cls.__name__)
        return {key: value}

    @classmethod
    def _decode_value(cls, buf, strict=True):
        readahead_tag, rest = tag.Tag.read(buf)
        length, rest = read_length(rest, strict=strict)
        if length == -1:
            if readahead_tag.encoding != tag.CONSTRUCTED:
                raise error.ASN1Error("Indefinite length encoding in primitive "
                                      "type")
            raise NotImplementedError("Indefinite length encoding of CHOICE "
                                      "type is not supported")

        if len(rest) != length:
            raise error.ASN1Error("Invalid length encoding")
        return cls._decode_readahead_value(buf, readahead_tag, rest,
                                           strict=strict)


class Repeated(Constructed, collections.MutableSequence):
    """Base class for SetOf and SequenceOf."""

    def __getitem__(self, index):
        return self._value[index]

    def __setitem__(self, index, value):
        # We are required to support both single-value as well as slice
        # assignment.
        if isinstance(index, slice):
            self._value[index] = self._convert_value(value)
        else:
            self._value[index] = (value if type(value) is self.component
                                  else self.component(value))
        self._modified = True

    def __delitem__(self, index):
        del self._value[index]
        self._modified = True

    def __len__(self):
        return len(self._value)

    def items(self):
        return enumerate(self._value)

    def insert(self, index, value):
        if type(value) is not self.component:
            value = self.component(value)
        self._value.insert(index, value)
        self._modified = True

    @property
    def value(self):
        return list(self._value)

    @classmethod
    def _convert_value(cls, value):
        return [x if type(x) is cls.component else cls.component(x)
                for x in value]


@Universal(16, tag.CONSTRUCTED)
class SequenceOf(Repeated):
    """Sequence Of."""

    def _encode_value(self):
        ret = [x.encode() for x in self._value]
        return "".join(ret)

    @classmethod
    def _decode_value(cls, buf, strict=True):
        ret = []
        while buf:
            value, buf = cls.component.read(buf, strict=strict)
            ret.append(value)
        return ret

    @classmethod
    def _read_indefinite_value(cls, buf):
        ret = []
        while len(buf) >= 2:
            if buf[:2] == _EOC:
                return ret, buf[2:]
            value, buf = cls.component.read(buf, strict=False)
            ret.append(value)
        raise error.ASN1Error("Missing EOC octets")


# We cannot use a real set to represent SetOf because
# (a) our components are mutable and thus not hashable and
# (b) ASN.1 allows duplicates: {1} and {1, 1} are distinct sets.
# Note that this means that eq-comparison is order-dependent.
@Universal(17, tag.CONSTRUCTED)
class SetOf(Repeated):
    """Set Of."""

    def _encode_value(self):
        ret = [x.encode() for x in self._value]
        ret.sort()
        return "".join(ret)

    @classmethod
    def _decode_value(cls, buf, strict=True):
        ret = []
        while buf:
            value, buf = cls.component.read(buf, strict=strict)
            ret.append(value)
        # TODO(ekasper): reject BER encodings in strict mode, i.e.,
        # verify sort order.
        return ret

    @classmethod
    def _read_indefinite_value(cls, buf):
        ret = []
        while len(buf) >= 2:
            if buf[:2] == _EOC:
                # TODO(ekasper): reject BER encodings in strict mode, i.e.,
                # verify sort order.
                return ret, buf[2:]
            value, buf = cls.component.read(buf, strict=False)
            ret.append(value)
        raise error.ASN1Error("Missing EOC octets")


class Component(object):
    """Sequence component specification."""

    def __init__(self, name, value_type, optional=False, default=None,
                 defined_by=None, lookup=None):
        """Define a sequence component.

        Args:
            name: component name. Must be unique within a sequence.
            value_type: the ASN.1 type.
            optional: if True, the component is optional.
            default: default value of the component.
            defined_by: for Any types, this specifies the component
                that defines the type.
            lookup: the lookup dictionary for Any types.
        """
        self.name = name
        self.value_type = value_type
        if default is None or type(default) is value_type:
            self.default = default
        else:
            self.default = value_type(default)
        if self.default is not None:
            self.encoded_default = self.default.encode()
        else:
            self.encoded_default = None
        self.optional = optional or (self.default is not None)
        self.defined_by = defined_by
        self.lookup = lookup


class MetaSequence(abc.ABCMeta):
    """Metaclass for building Sequence types."""

    def __new__(mcs, name, bases, dic):
        # Build a key -> component map for setting values.
        components = dic.get("components", ())
        if components:
            key_map = {}
            for component in components:
                if component.name in key_map:
                    raise TypeError("Duplicate name in Sequence specification")
                key_map[component.name] = component
            dic["key_map"] = key_map
        return super(MetaSequence, mcs).__new__(mcs, name, bases, dic)


@Universal(16, tag.CONSTRUCTED)
class Sequence(Constructed, collections.MutableMapping):
    """Sequence."""
    __metaclass__ = MetaSequence

    def __getitem__(self, key):
        return self._value[key]

    def __setitem__(self, key, value):
        component = self.key_map[key]
        value = self._convert_single_value(component, value)
        self._value[key] = value
        self._modified = True

    def __delitem__(self, key):
        if key not in self.key_map:
            raise KeyError("Invalid key %s" % key)
        self[key] = None
        self._modified = True

    def __iter__(self):
        """Iterate component names in order."""
        for component in self.components:
            yield component.name

    def __len__(self):
        """Missing optional components are counted in the length."""
        return len(self.components)

    @property
    def value(self):
        # Note that this does not preserve the component order.
        # However an order is encoded in the type spec, so we can still
        # recreate the original object from this value.
        return dict(self._value)

    def _encode_value(self):
        ret = []
        for component in self.components:
            value = self._value[component.name]
            if value is None:
                if not component.optional:
                    raise error.ASN1Error("Missing %s value in %s" %
                                          (component.name,
                                           self.__class__.__name__))
            else:
                # Value is not None.
                # We could compare by value for most types, but for "set" types
                # different values may yield the same encoding, so we compare
                # directly by encoding.
                # (Even though I haven't seen a defaulted set type in practice.)
                encoded_value = value.encode()
                if component.encoded_default != encoded_value:
                    ret.append(encoded_value)
        return "".join(ret)

    @classmethod
    def _convert_single_value(cls, component, value):
        # If value is None, we store the default if it is different from None.
        if value is None:
            return component.default
        elif type(value) is component.value_type:
            return value
        # If the supplied value is not of the exact same type then we discard
        # the tag information and try to construct from scratch.
        else:
            # TODO(ekasper): verify defined_by constraints here.
            return component.value_type(value)

    @classmethod
    def _convert_value(cls, value):
        ret = {}
        value = value or {}
        if not all([key in cls.key_map for key in value]):
            raise ValueError("Invalid keys in initializer")
        for component in cls.components:
            ret[component.name] = cls._convert_single_value(
                component, value.get(component.name, None))
        return ret

    @classmethod
    def _read_value(cls, buf, strict=True):
        ret = dict()
        for component in cls.components:
            try:
                value, buf = component.value_type.read(buf, strict=strict)
            except error.ASN1TagError:
                # If the component was optional and we got a tag mismatch,
                # assume decoding failed because the component was missing,
                # and carry on.
                # TODO(ekasper): since we let errors fall through recursively,
                # not all of the tag errors can be reasonably explained by
                # missing optional components. We could tighten this to match by
                # outermost tag only, and have metaclass verify the uniqueness
                # of component tags. Meanwhile, the worst that can happen is
                # that we retry in vain and don't return the most helpful error
                # message when we do finally fail.
                if not component.optional:
                    raise
                else:
                    ret[component.name] = component.default
            else:
                ret[component.name] = value

        # Second pass for decoding ANY.
        for component in cls.components:
            if component.defined_by is not None:
                value_type = component.lookup.get(
                    ret[component.defined_by], None)
                if value_type is not None:
                    try:
                        ret[component.name].decode_inner(value_type,
                                                         strict=strict)
                    except error.ASN1Error:
                        if strict:
                            raise
        return ret, buf

    @classmethod
    def _decode_value(cls, buf, strict=True):
        ret, buf = cls._read_value(buf, strict=strict)
        if buf:
            raise error.ASN1Error("Invalid encoding")
        return ret

    @classmethod
    def _read_indefinite_value(cls, buf):
        # We must be in strict=False mode by definition.
        ret, buf = cls._read_value(buf, strict=False)
        if buf[:2] != _EOC:
            raise error.ASN1Error("Missing EOC octets")
        return ret, buf[2:]



import six
from smime.crypto.asn1 import oid, types, tag
from smime.crypto.asn1 import x509_common
from smime.crypto.asn1 import x509_name
from smime.crypto.asn1 import x509_time
from smime.crypto.asn1 import x509_extension
from smime.crypto.asn1 import x509


# -- 10.1. Algorithm Identifier Types

_ALGORITHM_IDENTIFIER_DICT = {
    oid.RSA_ENCRYPTION: types.Null,
    oid.ID_AES128_CBC: types.OctetString,
    oid.ID_AES192_CBC: types.OctetString,
    oid.ID_AES256_CBC: types.OctetString
}

class AlgorithmIdentifier(types.Sequence):
    components = (
        (types.Component("algorithm", oid.ObjectIdentifier)),
        (types.Component("parameters", types.Any, optional=True,
                         defined_by="algorithm", lookup=_ALGORITHM_IDENTIFIER_DICT))
        )


class DigestAlgorithmIdentifier(AlgorithmIdentifier):
    """
    DigestAlgorithmIdentifier ::= AlgorithmIdentifier
    """
    pass


class SignatureAlgorithmIdentifier(AlgorithmIdentifier):
    """
    SignatureAlgorithmIdentifier ::= AlgorithmIdentifier
    """
    pass


class KeyEncryptionAlgorithmIdentifier(AlgorithmIdentifier):
    """
    KeyEncryptionAlgorithmIdentifier ::= AlgorithmIdentifier
    """
    pass


class ContentEncryptionAlgorithmIdentifier(AlgorithmIdentifier):
    """
    ContentEncryptionAlgorithmIdentifier ::= AlgorithmIdentifier
    """
    pass


class MessageAuthenticationCodeAlgorithm(AlgorithmIdentifier):
    """
    MessageAuthenticationCodeAlgorithm ::= AlgorithmIdentifier
    """
    pass


class KeyDerivationAlgorithmIdentifier(AlgorithmIdentifier):
    """
    KeyDerivationAlgorithmIdentifier ::= AlgorithmIdentifier
    """
    pass


# -- Preamble: General Definitions

class LabeledInteger(types.Integer):
    labels = ()

    def __str__(self):
        for k, v in self.labels:
            if v == self._value:
                return k
        return 'unk(%d)' % self._value

    @classmethod
    def _convert_value(self, value):
        if isinstance(value, six.string_types):
            for k, v in self.labels:
                if k == value:
                    return v
        return int(value)


@types.Universal(10, tag.PRIMITIVE)
class Enumerated(types.Simple):
    labels = ()

    def __str__(self):
        for k, v in self.labels:
            if v == self._value:
                return k
        return 'unk(%d)' % self._value

    @classmethod
    def _convert_value(self, value):
        if isinstance(value, six.string_types):
            for k, v in self.labels:
                if k == value:
                    return v
        raise TypeError("Cannot assign unknown label to enumerated type")

    def _encode_value(self):
        return types.encode_int(self._value)

    @classmethod
    def _decode_value(self, buf, strict=True):
        return types.decode_int(buf, strict=strict)


class ContentType(oid.ObjectIdentifier):
    pass


class CMSVersion(LabeledInteger):
    """
    CMSVersion ::= INTEGER  { v0(0), v1(1), v2(2), v3(3), v4(4) }
    """
    labels = (('v0', 0), ('v1', 1), ('v2', 2), ('v3', 3), ('v4', 4))


class SetOfAttributes(types.SetOf):
    print_labels = False
    print_delimiter = ", "
    component = x509_name.AttributeValue


# TODO: map X.509 attributes + ContentType, MessageDigest, SigningTime, Countersignature
class Attribute(types.Sequence):
    """
    Attribute ::= SEQUENCE {
        attrType OBJECT IDENTIFIER,
        attrValues SET OF AttributeValue }
    """
    components = (
        (types.Component('type', oid.ObjectIdentifier)),
        (types.Component('values', SetOfAttributes))
        )


class AuthAttributes(types.SetOf):
    """
    AuthAttributes ::= SET SIZE (1..MAX) OF Attribute
    """
    component = Attribute


class UnauthAttributes(types.SetOf):
    """
    UnauthAttributes ::= SET SIZE (1..MAX) OF Attribute
    """
    components = Attribute


class IssuerSerial(types.Sequence):
    """
    IssuerSerial  ::=  SEQUENCE {
       issuer         GeneralNames,
       serial         CertificateSerialNumber,
       issuerUID      UniqueIdentifier OPTIONAL
    }
    """
    components = (
        (types.Component('issuer', x509_name.GeneralNames)),
        (types.Component('serial', x509_common.CertificateSerialNumber)),
        (types.Component('issuerUID', x509_common.UniqueIdentifier, optional=True))
        )


class AttCertValidityPeriod(types.Sequence):
    """
    AttCertValidityPeriod ::= SEQUENCE {
         notBeforeTime  GeneralizedTime,
         notAfterTime   GeneralizedTime
    }
    """
    components = (
        (types.Component('notBeforeTime', x509_time.GeneralizedTime)),
        (types.Component('notAfterTime', x509_time.GeneralizedTime))
        )


class AttCertVersion(LabeledInteger):
    """
    AttCertVersion ::= INTEGER { v1(0), v2(1) }
    """
    labels = (('v1', 0), ('v2', 1))


class SequenceOfAttributes(types.SequenceOf):
    print_delimiter = ", "
    component = Attribute


class AttributeCertificateInfoV1(types.Sequence):
    """
    AttributeCertificateInfoV1 ::= SEQUENCE {
         version AttCertVersionV1 DEFAULT v1,
         subject CHOICE {
           baseCertificateID [0] IssuerSerial,
             -- associated with a Public Key Certificate
           subjectName [1] GeneralNames },
             -- associated with a name
         issuer GeneralNames,
         signature AlgorithmIdentifier,
         serialNumber CertificateSerialNumber,
         attCertValidityPeriod AttCertValidityPeriod,
         attributes SEQUENCE OF Attribute,
         issuerUniqueID UniqueIdentifier OPTIONAL,
         extensions Extensions OPTIONAL }
    """

    class SubjectChoices(types.Choice):
        print_labels = True
        components = {
            'baseCertificateID': IssuerSerial.implicit(0),
            'subjectName': x509_name.GeneralNames.implicit(1)
        }

    components = (
        (types.Component('version', AttCertVersion, default='v1')),
        (types.Component('subject', SubjectChoices)),
        (types.Component('issuer', x509_name.GeneralNames)),
        (types.Component('signature', x509_common.AlgorithmIdentifier)),
        (types.Component('serialNumber', x509_common.CertificateSerialNumber)),
        (types.Component('attCertValidityPeriod', AttCertValidityPeriod)),
        (types.Component('attributes', SequenceOfAttributes)),
        (types.Component('issuerUniqueID', x509_common.UniqueIdentifier, optional=True)),
        (types.Component('extensions', x509_extension.Extensions, optional=True))
        )


class AttributeCertificateV1(types.Sequence):
    """
    AttributeCertificateV1 ::= SEQUENCE {
         acInfo AttributeCertificateInfoV1,
         signatureAlgorithm AlgorithmIdentifier,
         signature BIT STRING }
    """
    components = (
        (types.Component('acInfo', AttributeCertificateInfoV1)),
        (types.Component('signatureAlgorithm', x509_common.AlgorithmIdentifier)),
        (types.Component('signature', types.BitString))
        )


class ObjectDigestInfo(types.Sequence):
    """
    ObjectDigestInfo ::= SEQUENCE {
         digestedObjectType  ENUMERATED {
                 publicKey            (0),
                 publicKeyCert        (1),
                 otherObjectTypes     (2) },
                         -- otherObjectTypes MUST NOT
                         -- be used in this profile
         otherObjectTypeID   OBJECT IDENTIFIER OPTIONAL,
         digestAlgorithm     AlgorithmIdentifier,
         objectDigest        BIT STRING
    }
    """
    class DigestedObjectType(Enumerated):
        values = (
            ('publicKey', 0),
            ('publicKeyCert', 1),
            ('otherObjectTypes', 2)
        )

    components = (
        (types.Component('digestObjectType', DigestedObjectType)),
        (types.Component('otherObjectTypeID', oid.ObjectIdentifier, optional=True)),
        (types.Component('digestAlgorithm', x509_common.AlgorithmIdentifier)),
        (types.Component('objectDigest', types.BitString))
        )


class Holder(types.Sequence):
    """
    Holder ::= SEQUENCE {
          baseCertificateID   [0] IssuerSerial OPTIONAL,
                   -- the issuer and serial number of
                   -- the holder's Public Key Certificate
          entityName          [1] GeneralNames OPTIONAL,
                   -- the name of the claimant or role
          objectDigestInfo    [2] ObjectDigestInfo OPTIONAL
                   -- used to directly authenticate the holder,
                   -- for example, an executable
    }
    """
    components = (
        (types.Component('baseCertificateID', IssuerSerial.implicit(0), optional=True)),
        (types.Component('entityName', x509_name.GeneralNames.implicit(1), optional=True)),
        (types.Component('objectDigestInfo', ObjectDigestInfo.implicit(2), optional=True))
        )


class V2Form(types.Sequence):
    """
    V2Form ::= SEQUENCE {
         issuerName            GeneralNames  OPTIONAL,
         baseCertificateID     [0] IssuerSerial  OPTIONAL,
         objectDigestInfo      [1] ObjectDigestInfo  OPTIONAL
           -- issuerName MUST be present in this profile
           -- baseCertificateID and objectDigestInfo MUST NOT
           -- be present in this profile
    }
    """
    components = (
        (types.Component('issuerName', x509_name.GeneralNames, optional=True)),
        (types.Component('baseCertificateID', IssuerSerial.implicit(0), optional=True)),
        (types.Component('objectDigestInfo', ObjectDigestInfo.implicit(1), optional=True))
        )


class AttCertIssuer(types.Choice):
    """
    AttCertIssuer ::= CHOICE {
         v1Form   GeneralNames,  -- MUST NOT be used in this
                                 -- profile
         v2Form   [0] V2Form     -- v2 only
    }
    """
    print_labels = True
    components = {
        'v1Form': x509_name.GeneralNames,
        'v2Form': V2Form.implicit(0)
        }


class AttributeCertificateInfo(types.Sequence):
    """
    AttributeCertificateInfo ::= SEQUENCE {
         version              AttCertVersion -- version is v2,
         holder               Holder,
         issuer               AttCertIssuer,
         signature            AlgorithmIdentifier,
         serialNumber         CertificateSerialNumber,
         attrCertValidityPeriod   AttCertValidityPeriod,
         attributes           SEQUENCE OF Attribute,
         issuerUniqueID       UniqueIdentifier OPTIONAL,
         extensions           Extensions OPTIONAL
    }
    """
    components = (
        (types.Component('version', AttCertVersion, default='v2')),
        (types.Component('holder', Holder)),
        (types.Component('issuer', AttCertIssuer)),
        (types.Component('signature', x509_common.AlgorithmIdentifier)),
        (types.Component('serialNumber', x509_common.CertificateSerialNumber)),
        (types.Component('attrCertValidityPeriod', AttCertValidityPeriod)),
        (types.Component('attributes', SequenceOfAttributes)),
        (types.Component('issuerUniqueID', x509_common.UniqueIdentifier, optional=True)),
        (types.Component('extensions', x509_extension.Extensions, optional=True))
        )


class AttributeCertificate(types.Sequence):
    """
    AttributeCertificate ::= SEQUENCE {
        acinfo               AttributeCertificateInfo,
        signatureAlgorithm   AlgorithmIdentifier,
        signatureValue       BIT STRING
    }
    """
    components = (
        (types.Component('acinfo', AttributeCertificateInfo)),
        (types.Component('signatureAlgorithm', x509_common.AlgorithmIdentifier)),
        (types.Component('signatureValue', types.BitString))
        )


class AttributeCertificateV2(AttributeCertificate):
    """
    AttributeCertificateV2 ::= AttributeCertificate
    """
    pass


class ExtendedCertificateInfo(types.Sequence):
    """
    ExtendedCertificateInfo ::= SEQUENCE {
         version CMSVersion,
         certificate Certificate,
         attributes UnauthAttributes }
    """
    components = (
        (types.Component('version', CMSVersion)),
        (types.Component('certificate', x509.Certificate)),
        (types.Component('attributes', UnauthAttributes))
        )


class Signature(types.BitString):
    pass


class ExtendedCertificate(types.Sequence):
    """
    ExtendedCertificate ::= SEQUENCE {
         extendedCertificateInfo ExtendedCertificateInfo,
         signatureAlgorithm SignatureAlgorithmIdentifier,
         signature Signature }
    """
    components = (
        (types.Component('extendedCertificateInfo', ExtendedCertificateInfo)),
        (types.Component('signatureAlgorithm', SignatureAlgorithmIdentifier)),
        (types.Component('signature', Signature))
        )


class CertificateChoices(types.Choice):
    """
    CertificateChoices ::= CHOICE {
      certificate Certificate,
      extendedCertificate [0] IMPLICIT ExtendedCertificate,  -- Obsolete
      v1AttrCert [1] IMPLICIT AttributeCertificateV1,        -- Obsolete
      v2AttrCert [2] IMPLICIT AttributeCertificateV2 }
    """
    print_labels = True
    components = {
        'certificate': x509.Certificate,
        'extendedCertificate': ExtendedCertificate.implicit(0),
        'v1AttrCert': AttributeCertificateV1.implicit(1),
        'v2AttrCert': AttributeCertificateV2.implicit(2)
        }


class CertificateSet(types.SetOf):
    """
    CertificateSet ::= SET OF CertificateChoices
    """
    print_labels = False
    print_delimiter = ", "
    component = CertificateChoices


class EncapsulatedContentInfo(types.Sequence):
    """
    EncapsulatedContentInfo ::= SEQUENCE {
      eContentType ContentType,
      eContent [0] EXPLICIT OCTET STRING OPTIONAL }
    """
    components = (
        (types.Component('eContentType', ContentType)),
        (types.Component('eContent', types.OctetString.implicit(0), optional=True))
        )


class SubjectKeyIdentifier(types.OctetString):
    """
    SubjectKeyIdentifier ::= OCTET STRING
    """
    pass


class IssuerAndSerialNumber(types.Sequence):
    components = (
        (types.Component('issuer', x509_name.Name)),
        (types.Component('serialNumber', x509_common.CertificateSerialNumber))
        )


class SignerIdentifier(types.Choice):
    """
    SignerIdentifier ::= CHOICE {
        issuerAndSerialNumber IssuerAndSerialNumber,
        subjectKeyIdentifier [0] SubjectKeyIdentifier }
    """
    print_labels = True
    components = {
        'issuerAndSerialNumber': IssuerAndSerialNumber,
        'subjectKeyIdentifier': SubjectKeyIdentifier.implicit(0)
        }


class SignedAttributes(types.SetOf):
    """
    SignedAttributes ::= SET SIZE (1..MAX) OF Attribute
    """
    component = Attribute


class UnsignedAttributes(types.SetOf):
    """
    UnsignedAttributes ::= SET SIZE (1..MAX) OF Attribute
    """
    component = Attribute


class SignatureValue(types.OctetString):
    """
    SignatureValue ::= OCTET STRING
    """
    pass


class SignerInfo(types.Sequence):
    """
    SignerInfo ::= SEQUENCE {
        version CMSVersion,
        sid SignerIdentifier,
        digestAlgorithm DigestAlgorithmIdentifier,
        signedAttrs [0] IMPLICIT SignedAttributes OPTIONAL,
        signatureAlgorithm SignatureAlgorithmIdentifier,
        signature SignatureValue,
        unsignedAttrs [1] IMPLICIT UnsignedAttributes OPTIONAL }
    """
    components = (
        (types.Component('version', CMSVersion)),
        (types.Component('sid', SignerIdentifier)),
        (types.Component('digestAlgorithm', DigestAlgorithmIdentifier)),
        (types.Component('signedAttrs', SignedAttributes.implicit(0), optional=True)),
        (types.Component('signatureAlgorithm', SignatureAlgorithmIdentifier)),
        (types.Component('signature', SignatureValue)),
        (types.Component('unsignedAttrs', UnsignedAttributes.implicit(1), optional=True))
        )


class SignerInfos(types.SetOf):
    component = SignerInfo


class DigestAlgorithmIdentifiers(types.SetOf):
    component = DigestAlgorithmIdentifier


class SerialNumber(types.Integer):
    pass


class CRLEntry(types.Sequence):
    components = (
        (types.Component('userCertificate', SerialNumber)),
        (types.Component('revocationDate', x509_time.UTCTime))
        )


class CRLEntries(types.SequenceOf):
    component = CRLEntry


class TBSCertificateRevocationList(types.Sequence):
    components = (
        (types.Component('signature', x509_common.AlgorithmIdentifier)),
        (types.Component('issuer', x509_name.Name)),
        (types.Component('lastUpdate', x509_time.UTCTime)),
        (types.Component('nextUpdate', x509_time.UTCTime)),
        (types.Component('revokedCertificates', CRLEntries, optional=True))
        )


class CertificateRevocationList(types.Sequence):
    componentType = (
        (types.Component('tbsCertificateRevocationList', TBSCertificateRevocationList)),
        (types.Component('signatureAlgorithm', x509_common.AlgorithmIdentifier)),
        (types.Component('signature', types.BitString))
        )


class CertificateRevocationLists(types.SetOf):
    component = CertificateRevocationList


class OriginatorInfo(types.Sequence):
    """
    OriginatorInfo ::= SEQUENCE {
      certs [0] IMPLICIT CertificateSet OPTIONAL,
      crls [1] IMPLICIT CertificateRevocationLists OPTIONAL }
    """
    components = (
        (types.Component('certs', CertificateSet.implicit(0), optional=True)),
        (types.Component('crls', CertificateRevocationLists.implicit(1), optional=True)),
        )


class UnprotectedAttributes(types.SetOf):
    """
    UnprotectedAttributes ::= SET SIZE (1..MAX) OF Attribute
    """
    component = Attribute


class RecipientIdentifier(types.Choice):
    """
    RecipientIdentifier ::= CHOICE {
      issuerAndSerialNumber IssuerAndSerialNumber,
      subjectKeyIdentifier [0] SubjectKeyIdentifier }
    """
    print_labels = True
    components = {
        'issuerAndSerialNumber': IssuerAndSerialNumber,
        'subjectKeyIdentifier': SubjectKeyIdentifier.implicit(0)
        }


class EncryptedKey(types.OctetString):
    pass


class KeyTransRecipientInfo(types.Sequence):
    """
    KeyTransRecipientInfo ::= SEQUENCE {
      version CMSVersion,  -- always set to 0 or 2
      rid RecipientIdentifier,
      keyEncryptionAlgorithm KeyEncryptionAlgorithmIdentifier,
      encryptedKey EncryptedKey }
    """
    components = (
        (types.Component('version', CMSVersion)),
        (types.Component('rid', RecipientIdentifier)),
        (types.Component('keyEncryptionAlgorithm', KeyEncryptionAlgorithmIdentifier)),
        (types.Component('encryptedKey', EncryptedKey))
        )


class OriginatorPublicKey(types.Sequence):
    """
    OriginatorPublicKey ::= SEQUENCE {
      algorithm AlgorithmIdentifier,
      publicKey BIT STRING }
    """
    components = (
        (types.Component('algorithm', x509_common.AlgorithmIdentifier)),
        (types.Component('publicKey', types.BitString))
        )


class OriginatorIdentifierOrKey(types.Choice):
    """
    OriginatorIdentifierOrKey ::= CHOICE {
      issuerAndSerialNumber IssuerAndSerialNumber,
      subjectKeyIdentifier [0] SubjectKeyIdentifier,
      originatorKey [1] OriginatorPublicKey }
    """
    print_labels = True
    components = {
        'issuerAndSerialNumber': IssuerAndSerialNumber,
        'subjectKeyIdentifier': SubjectKeyIdentifier.implicit(0),
        'originatorKey': OriginatorPublicKey.implicit(1)
        }


class UserKeyingMaterial(types.OctetString):
    """
    UserKeyingMaterial ::= OCTET STRING
    """
    pass


class OtherKeyAttribute(types.Sequence):
    """
    OtherKeyAttribute ::= SEQUENCE {
      keyAttrId OBJECT IDENTIFIER,
      keyAttr ANY DEFINED BY keyAttrId OPTIONAL }
    """
    components = (
        (types.Component('keyAttrId', oid.ObjectIdentifier)),
        (types.Component('keyAttr', types.Any, optional=True))
        )


class RecipientKeyIdentifier(types.Sequence):
    """
    RecipientKeyIdentifier ::= SEQUENCE {
      subjectKeyIdentifier SubjectKeyIdentifier,
      date GeneralizedTime OPTIONAL,
      other OtherKeyAttribute OPTIONAL }
    """
    components = (
        (types.Component('subjectKeyIdentifier', SubjectKeyIdentifier)),
        (types.Component('date', x509_time.GeneralizedTime, optional=True)),
        (types.Component('other', OtherKeyAttribute, optional=True))
        )


class KeyAgreeRecipientIdentifier(types.Choice):
    """
    KeyAgreeRecipientIdentifier ::= CHOICE {
      issuerAndSerialNumber IssuerAndSerialNumber,
      rKeyId [0] IMPLICIT RecipientKeyIdentifier }
    """
    print_labels = True
    components = {
        'issuerAndSerialNumber': IssuerAndSerialNumber,
        'rKeyId': RecipientKeyIdentifier.implicit(0)
        }


class RecipientEncryptedKey(types.Sequence):
    """
    RecipientEncryptedKey ::= SEQUENCE {
      rid KeyAgreeRecipientIdentifier,
      encryptedKey EncryptedKey }
    """
    components = (
        (types.Component('rid', KeyAgreeRecipientIdentifier)),
        (types.Component('encryptedKey', EncryptedKey))
        )


class RecipientEncryptedKeys(types.SequenceOf):
    """
    RecipientEncryptedKeys ::= SEQUENCE OF RecipientEncryptedKey
    """
    component = RecipientEncryptedKey


class KeyAgreeRecipientInfo(types.Sequence):
    """
    KeyAgreeRecipientInfo ::= SEQUENCE {
      version CMSVersion,  -- always set to 3
      originator [0] EXPLICIT OriginatorIdentifierOrKey,
      ukm [1] EXPLICIT UserKeyingMaterial OPTIONAL,
      keyEncryptionAlgorithm KeyEncryptionAlgorithmIdentifier,
      recipientEncryptedKeys RecipientEncryptedKeys }
    """
    components = (
        (types.Component('version', CMSVersion)),
        (types.Component('originator', OriginatorIdentifierOrKey.explicit(0))),
        (types.Component('ukm', UserKeyingMaterial.explicit(1), optional=True)),
        (types.Component('keyEncryptionAlgorithm', KeyEncryptionAlgorithmIdentifier)),
        (types.Component('recipientEncryptedKeys', RecipientEncryptedKeys))
        )


class KEKIdentifier(types.Sequence):
    """
    KEKIdentifier ::= SEQUENCE {
      keyIdentifier OCTET STRING,
      date GeneralizedTime OPTIONAL,
      other OtherKeyAttribute OPTIONAL }
    """
    components = (
        (types.Component('keyIdentifier', types.OctetString)),
        (types.Component('date', x509_time.GeneralizedTime, optional=True)),
        (types.Component('other', OtherKeyAttribute, optional=True))
        )


class KEKRecipientInfo(types.Sequence):
    """
    KEKRecipientInfo ::= SEQUENCE {
      version CMSVersion,  -- always set to 4
      kekid KEKIdentifier,
      keyEncryptionAlgorithm KeyEncryptionAlgorithmIdentifier,
      encryptedKey EncryptedKey }
    """
    components = (
        (types.Component('version', CMSVersion)),
        (types.Component('kekid', KEKIdentifier)),
        (types.Component('keyEncryptionAlgorithm', KeyEncryptionAlgorithmIdentifier)),
        (types.Component('encryptedKey', EncryptedKey))
        )


class PasswordRecipientInfo(types.Sequence):
    """
    PasswordRecipientInfo ::= SEQUENCE {
      version CMSVersion,   -- always set to 0
      keyDerivationAlgorithm [0] KeyDerivationAlgorithmIdentifier
                                 OPTIONAL,
      keyEncryptionAlgorithm KeyEncryptionAlgorithmIdentifier,
      encryptedKey EncryptedKey }
    """
    components = (
        (types.Component('version', CMSVersion)),
        (types.Component('keyDerivationAlgorithm', KeyDerivationAlgorithmIdentifier.implicit(0), optional=True)),
        (types.Component('keyEncryptionAlgorithm', KeyEncryptionAlgorithmIdentifier)),
        (types.Component('encryptedKey', EncryptedKey))
        )


class OtherRecipientInfo(types.Sequence):
    """
    OtherRecipientInfo ::= SEQUENCE {
         oriType OBJECT IDENTIFIER,
         oriValue ANY DEFINED BY oriType }
    """
    components = (
        (types.Component('oriType', oid.ObjectIdentifier)),
        (types.Component('oriValue', types.Any))
        )


class RecipientInfo(types.Choice):
    """
    RecipientInfo ::= CHOICE {
         ktri KeyTransRecipientInfo,
         kari [1] KeyAgreeRecipientInfo,
         kekri [2] KEKRecipientInfo,
         pwri [3] PasswordRecipientInfo,
         ori [4] OtherRecipientInfo }
    """
    print_labels = True
    components = {
        'ktri': KeyTransRecipientInfo,
        'kari': KeyAgreeRecipientInfo.implicit(1),
        'kekri': KEKRecipientInfo.implicit(2),
        'pwri': PasswordRecipientInfo.implicit(3),
        'ori': OtherRecipientInfo.implicit(4)
        }


class RecipientInfos(types.SetOf):
    """
    RecipientInfos ::= SET SIZE (1..MAX) OF RecipientInfo
    """
    component = RecipientInfo


class EncryptedContent(types.OctetString):
    pass


class EncryptedContentInfo(types.Sequence):
    components = (
        (types.Component('contentType', ContentType)),
        (types.Component('contentEncryptionAlgorithm', ContentEncryptionAlgorithmIdentifier)),
        (types.Component('encryptedContent', EncryptedContent.implicit(0), optional=True))
        )


class Digest(types.OctetString):
    pass


class MessageAuthenticationCode(types.OctetString):
    """
    MessageAuthenticationCode ::= OCTET STRING
    """
    pass



"""Named values."""


class NamedValue(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __repr__(self):
        return "%r(%s, %s)" % (self.__class__.__name__, self.name, self.value)

    def __str__(self):
        return self.name



# Refer to RFC3565

import base64
import Crypto.Random.OSRNG as RNG
from email import message_from_string
from email.mime.text import MIMEText

from .rsa import RSAPublicKey
from .aes import AES128_CBC, AES192_CBC, AES256_CBC

from smime.crypto.asn1 import oid, types
from smime.crypto.asn1 import cms
from smime.crypto.asn1 import cms_common


def __get_issuer_and_serial_number(x509_cert):
    tbsCertificate = x509_cert['tbsCertificate']
    return cms_common.IssuerAndSerialNumber({
        'issuer': tbsCertificate['issuer'],
        'serialNumber': tbsCertificate['serialNumber']
    })


def __get_enveloped_data(pubkey_cipher, sym_cipher, x509_cert, encrypted_key, iv, encrypted_content):
    return cms.ContentInfo({
        'contentType': oid.ID_ENVELOPED_DATA,
        'content': cms.EnvelopedData({
            'version': cms_common.CMSVersion('v0'),
            'recipientInfos': cms_common.RecipientInfos([
                cms_common.RecipientInfo({
                    'ktri': cms_common.KeyTransRecipientInfo({
                        'version': cms_common.CMSVersion('v0'),
                        'rid': cms_common.RecipientIdentifier({
                            'issuerAndSerialNumber':
                                __get_issuer_and_serial_number(x509_cert)
                        }),
                        'keyEncryptionAlgorithm':
                            cms_common.KeyEncryptionAlgorithmIdentifier({
                                'algorithm': pubkey_cipher.oid,
                                'parameters': types.Null(False)
                            }),
                        'encryptedKey':  cms_common.EncryptedKey(encrypted_key)
                    })
                })
            ]),
            'encryptedContentInfo': cms_common.EncryptedContentInfo({
                'contentType': oid.ID_DATA,
                'contentEncryptionAlgorithm': cms_common.ContentEncryptionAlgorithmIdentifier({
                    'algorithm': sym_cipher.oid,
                    'parameters': types.OctetString(iv)
                }),
                'encryptedContent': cms_common.EncryptedContent(encrypted_content)
            })
        })
    })


def __encode_in_base64(stream):
    columns = 64
    result = []
    stream = base64.b64encode(stream)
    while len(stream) > 0:
        chunk = stream[:columns]
        stream = stream[columns:]
        result.append(chunk)
    return '\n'.join(result)


def __encrypt_internal(rsa, algo, content, session_key):
    """
    Takes the contents of the message parameter, formatted as in RFC 2822, and encrypts them,
    so that they can only be read by the intended recipient specified by pubkey.
    :return: string containing the new encrypted message.
    """
    return rsa.encrypt(session_key), algo.encrypt(content)


def encrypt(message, pubkey, algorithm='aes256'):
    """
    Takes the contents of the message parameter, formatted as in RFC 2822, and encrypts them,
    so that they can only be read by the intended recipient specified by pubkey.
    :return: string containing the new encrypted message.
    """

    if algorithm == 'aes256':
        Algo = AES256_CBC
    elif algorithm == 'aes192':
        Algo = AES192_CBC
    elif algorithm == 'aes128':
        Algo = AES128_CBC
    else:
        raise ValueError('Unknown algorithm')

    # Get the message content
    msg = message_from_string(message)
    to_encode = MIMEText(msg.get_payload())
    content = to_encode.as_string()

    pubkey_cipher = RSAPublicKey(pubkey)

    session_key = RNG.new().read(Algo.key_size)
    iv = RNG.new().read(Algo.block_size)
    sym_cipher = Algo(session_key, iv)

    encrypted_key = pubkey_cipher.encrypt(session_key)
    encrypted_content = sym_cipher.encrypt(content)

    # Encode the content
    enveloped_data = __get_enveloped_data(pubkey_cipher, sym_cipher,
            pubkey_cipher.get_cert(), encrypted_key, iv, encrypted_content)
    encoded_content = __encode_in_base64(enveloped_data.encode())

    # Create the resulting message
    result_msg = MIMEText(encoded_content)
    overrides = (
        ('MIME-Version', '1.0'),
        ('Content-Type', 'application/pkcs7-mime; smime-type=enveloped-data; name=smime.p7m'),
        ('Content-Transfer-Encoding', 'base64'),
        ('Content-Disposition', 'attachment; filename=smime.p7m')
    )

    for name, value in list(msg.items()):
        if name in [x for x, _ in  overrides]:
            continue
        result_msg.add_header(name, value)

    for name, value in overrides:
        if name in result_msg:
            del result_msg[name]
        result_msg[name] = value

    return result_msg.as_string()



#!/usr/bin/env python

import unittest

from smime.crypto import error
from smime.crypto.asn1 import x509_time


class TimeTest(unittest.TestCase):
    def verify_time(self, time_struct, year, month, day, hour, minute, sec):
        self.assertEqual(year, time_struct.tm_year)
        self.assertEqual(month, time_struct.tm_mon)
        self.assertEqual(day, time_struct.tm_mday)
        self.assertEqual(hour, time_struct.tm_hour)
        self.assertEqual(minute, time_struct.tm_min)
        self.assertEqual(sec, time_struct.tm_sec)

    def test_time(self):
        t = x509_time.UTCTime(value="130822153902Z").gmtime()
        self.verify_time(t, 2013, 8, 22, 15, 39, 2)

        t = x509_time.GeneralizedTime(value="20130822153902Z").gmtime()
        self.verify_time(t, 2013, 8, 22, 15, 39, 2)

    def test_utc_time_1900(self):
        t = x509_time.UTCTime(value="500822153902Z").gmtime()
        self.verify_time(t, 1950, 8, 22, 15, 39, 2)

    def test_time_invalid(self):
        self.assertRaises(error.ASN1Error, x509_time.UTCTime,
                          value="131322153902Z")
        self.assertRaises(error.ASN1Error, x509_time.UTCTime,
                          value="201301322153902Z")
        t = x509_time.UTCTime(value="131322153902Z", strict=False)
        self.assertRaises(error.ASN1Error, t.gmtime)
        t = x509_time.UTCTime(value="201301322153902Z", strict=False)
        self.assertRaises(error.ASN1Error, t.gmtime)

    def test_time_no_seconds(self):
        t = x509_time.UTCTime(value="0001010000Z").gmtime()
        self.verify_time(t, 2000, 1, 1, 0, 0, 0)

    def test_time_alt_gmt(self):
        t = x509_time.UTCTime(value="121214093107+0000").gmtime()
        self.verify_time(t, 2012, 12, 14, 9, 31, 7)

    def test_time_alt_tz(self):
        """
        Test parsing a timezone with old +HHMM offset format
        Right now, it is ignored.
        """
        t = x509_time.UTCTime(value="121214093107+1234").gmtime()
        self.verify_time(t, 2012, 12, 14, 9, 31, 7)

    def test_time_missing_z(self):
        self.assertRaises(x509_time.UTCTime, value="130822153902", strict=True)

        t2 = x509_time.UTCTime(value="130822153902", strict=False).gmtime()
        self.verify_time(t2, 2013, 8, 22, 15, 39, 2)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

import unittest

from smime.crypto.asn1 import tag


class TagTest(unittest.TestCase):
    """Test tag encoding."""

    def test_encode_read(self):
        valid_tags = (
            # (initializers, encoding)
            ((0, tag.UNIVERSAL, tag.PRIMITIVE), "\x00"),
            ((1, tag.UNIVERSAL, tag.PRIMITIVE), "\x01"),
            ((16256, tag.UNIVERSAL, tag.PRIMITIVE), "\x1f\xff\x00"),
            ((16, tag.UNIVERSAL, tag.CONSTRUCTED), "\x30"),
            ((17, tag.UNIVERSAL, tag.CONSTRUCTED), "\x31"),
            ((16256, tag.UNIVERSAL, tag.CONSTRUCTED), "\x3f\xff\x00"),
            ((0, tag.APPLICATION, tag.PRIMITIVE), "\x40"),
            ((1, tag.APPLICATION, tag.PRIMITIVE), "\x41"),
            ((16256, tag.APPLICATION, tag.PRIMITIVE), "\x5f\xff\x00"),
            ((0, tag.APPLICATION, tag.CONSTRUCTED), "\x60"),
            ((1, tag.APPLICATION, tag.CONSTRUCTED), "\x61"),
            ((16256, tag.APPLICATION, tag.CONSTRUCTED), "\x7f\xff\x00"),
            ((0, tag.CONTEXT_SPECIFIC, tag.PRIMITIVE), "\x80"),
            ((1, tag.CONTEXT_SPECIFIC, tag.PRIMITIVE), "\x81"),
            ((16256, tag.CONTEXT_SPECIFIC, tag.PRIMITIVE), "\x9f\xff\x00"),
            ((0, tag.CONTEXT_SPECIFIC, tag.CONSTRUCTED), "\xa0"),
            ((1, tag.CONTEXT_SPECIFIC, tag.CONSTRUCTED), "\xa1"),
            ((16256, tag.CONTEXT_SPECIFIC, tag.CONSTRUCTED), "\xbf\xff\x00"),
            ((0, tag.PRIVATE, tag.PRIMITIVE), "\xc0"),
            ((1, tag.PRIVATE, tag.PRIMITIVE), "\xc1"),
            ((16256, tag.PRIVATE, tag.PRIMITIVE), "\xdf\xff\x00"),
            ((0, tag.PRIVATE, tag.CONSTRUCTED), "\xe0"),
            ((1, tag.PRIVATE, tag.CONSTRUCTED), "\xe1"),
            ((16256, tag.PRIVATE, tag.CONSTRUCTED), "\xff\xff\x00"),
            )

        for init, enc in valid_tags:
            number, tag_class, encoding = init
            t = tag.Tag(number, tag_class, encoding)
            self.assertEqual(t.number, number)
            self.assertEqual(t.tag_class, tag_class)
            self.assertEqual(t.encoding, encoding)
            self.assertEqual(t.value, enc)
            self.assertEqual((t, ""), tag.Tag.read(enc))
            self.assertEqual((t, "rest"), tag.Tag.read(enc + "rest"))

        for i in range(len(valid_tags)):
            for j in range(i+1, len(valid_tags)):
                self.assertNotEqual(tag.Tag(*valid_tags[i][0]),
                                    tag.Tag(*valid_tags[j][0]))


if __name__ == '__main__':
    unittest.main()
